# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/editors/mainEditor_ui.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_mainEditor(object):
    def setupUi(self, mainEditor):
        mainEditor.setObjectName("mainEditor")
        mainEditor.resize(791, 319)
        self.verticalLayout = QtWidgets.QVBoxLayout(mainEditor)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabSplitter = tabSplitter(mainEditor)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabSplitter.sizePolicy().hasHeightForWidth())
        self.tabSplitter.setSizePolicy(sizePolicy)
        self.tabSplitter.setObjectName("tabSplitter")
        self.verticalLayout.addWidget(self.tabSplitter)
        self.horizontalLayout_19 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_19.setObjectName("horizontalLayout_19")
        self.btnGoUp = QtWidgets.QPushButton(mainEditor)
        self.btnGoUp.setText("")
        icon = QtGui.QIcon.fromTheme("go-up")
        self.btnGoUp.setIcon(icon)
        self.btnGoUp.setFlat(True)
        self.btnGoUp.setObjectName("btnGoUp")
        self.horizontalLayout_19.addWidget(self.btnGoUp)
        self.btnRedacFolderText = QtWidgets.QPushButton(mainEditor)
        self.btnRedacFolderText.setCheckable(True)
        self.btnRedacFolderText.setFlat(True)
        self.btnRedacFolderText.setObjectName("btnRedacFolderText")
        self.buttonGroup = QtWidgets.QButtonGroup(mainEditor)
        self.buttonGroup.setObjectName("buttonGroup")
        self.buttonGroup.addButton(self.btnRedacFolderText)
        self.horizontalLayout_19.addWidget(self.btnRedacFolderText)
        self.btnRedacFolderCork = QtWidgets.QPushButton(mainEditor)
        self.btnRedacFolderCork.setCheckable(True)
        self.btnRedacFolderCork.setChecked(True)
        self.btnRedacFolderCork.setFlat(True)
        self.btnRedacFolderCork.setObjectName("btnRedacFolderCork")
        self.buttonGroup.addButton(self.btnRedacFolderCork)
        self.horizontalLayout_19.addWidget(self.btnRedacFolderCork)
        self.btnRedacFolderOutline = QtWidgets.QPushButton(mainEditor)
        self.btnRedacFolderOutline.setCheckable(True)
        self.btnRedacFolderOutline.setFlat(True)
        self.btnRedacFolderOutline.setObjectName("btnRedacFolderOutline")
        self.buttonGroup.addButton(self.btnRedacFolderOutline)
        self.horizontalLayout_19.addWidget(self.btnRedacFolderOutline)
        self.sldCorkSizeFactor = QtWidgets.QSlider(mainEditor)
        self.sldCorkSizeFactor.setMinimumSize(QtCore.QSize(100, 0))
        self.sldCorkSizeFactor.setMaximumSize(QtCore.QSize(200, 16777215))
        self.sldCorkSizeFactor.setMinimum(50)
        self.sldCorkSizeFactor.setMaximum(200)
        self.sldCorkSizeFactor.setProperty("value", 100)
        self.sldCorkSizeFactor.setOrientation(QtCore.Qt.Horizontal)
        self.sldCorkSizeFactor.setObjectName("sldCorkSizeFactor")
        self.horizontalLayout_19.addWidget(self.sldCorkSizeFactor)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_19.addItem(spacerItem)
        self.lblRedacWC = QtWidgets.QLabel(mainEditor)
        self.lblRedacWC.setMinimumSize(QtCore.QSize(10, 0))
        self.lblRedacWC.setText("")
        self.lblRedacWC.setObjectName("lblRedacWC")
        self.horizontalLayout_19.addWidget(self.lblRedacWC)
        self.lblRedacProgress = QtWidgets.QLabel(mainEditor)
        self.lblRedacProgress.setMinimumSize(QtCore.QSize(100, 6))
        self.lblRedacProgress.setMaximumSize(QtCore.QSize(200, 14))
        self.lblRedacProgress.setText("")
        self.lblRedacProgress.setObjectName("lblRedacProgress")
        self.horizontalLayout_19.addWidget(self.lblRedacProgress)
        self.btnRedacFullscreen = QtWidgets.QPushButton(mainEditor)
        self.btnRedacFullscreen.setText("")
        icon = QtGui.QIcon.fromTheme("view-fullscreen")
        self.btnRedacFullscreen.setIcon(icon)
        self.btnRedacFullscreen.setFlat(True)
        self.btnRedacFullscreen.setObjectName("btnRedacFullscreen")
        self.horizontalLayout_19.addWidget(self.btnRedacFullscreen)
        self.verticalLayout.addLayout(self.horizontalLayout_19)

        self.retranslateUi(mainEditor)
        QtCore.QMetaObject.connectSlotsByName(mainEditor)

    def retranslateUi(self, mainEditor):
        _translate = QtCore.QCoreApplication.translate
        mainEditor.setWindowTitle(_translate("mainEditor", "Form"))
        self.btnGoUp.setToolTip(_translate("mainEditor", "Go to parent item"))
        self.btnGoUp.setShortcut(_translate("mainEditor", "Alt+Up"))
        self.btnRedacFolderText.setText(_translate("mainEditor", "Text"))
        self.btnRedacFolderCork.setText(_translate("mainEditor", "Index cards"))
        self.btnRedacFolderOutline.setText(_translate("mainEditor", "Outline"))
        self.btnRedacFullscreen.setShortcut(_translate("mainEditor", "F11"))

from manuskript.ui.editors.tabSplitter import tabSplitter



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/cheatSheet_ui.ui'
#
# Created: Fri Apr  8 14:27:04 2016
#      by: PyQt5 UI code generator 5.2.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_cheatSheet(object):
    def setupUi(self, cheatSheet):
        cheatSheet.setObjectName("cheatSheet")
        cheatSheet.resize(400, 344)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(cheatSheet)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.txtFilter = QtWidgets.QLineEdit(cheatSheet)
        self.txtFilter.setStyleSheet("")
        self.txtFilter.setFrame(False)
        self.txtFilter.setClearButtonEnabled(True)
        self.txtFilter.setObjectName("txtFilter")
        self.horizontalLayout.addWidget(self.txtFilter)
        self.btnShowList = QtWidgets.QPushButton(cheatSheet)
        self.btnShowList.setText("")
        icon = QtGui.QIcon.fromTheme("go-bottom")
        self.btnShowList.setIcon(icon)
        self.btnShowList.setCheckable(True)
        self.btnShowList.setFlat(True)
        self.btnShowList.setObjectName("btnShowList")
        self.horizontalLayout.addWidget(self.btnShowList)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.line = QtWidgets.QFrame(cheatSheet)
        self.line.setFrameShadow(QtWidgets.QFrame.Plain)
        self.line.setLineWidth(0)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout_2.addWidget(self.line)
        self.splitter = QtWidgets.QSplitter(cheatSheet)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.list = QtWidgets.QListWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.list.sizePolicy().hasHeightForWidth())
        self.list.setSizePolicy(sizePolicy)
        self.list.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.list.setObjectName("list")
        self.scrollArea = QtWidgets.QScrollArea(self.splitter)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setLineWidth(0)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 400, 68))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.view = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.view.sizePolicy().hasHeightForWidth())
        self.view.setSizePolicy(sizePolicy)
        self.view.setLineWidth(0)
        self.view.setText("")
        self.view.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.view.setWordWrap(True)
        self.view.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.view.setObjectName("view")
        self.verticalLayout.addWidget(self.view)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout_2.addWidget(self.splitter)

        self.retranslateUi(cheatSheet)
        QtCore.QMetaObject.connectSlotsByName(cheatSheet)

    def retranslateUi(self, cheatSheet):
        _translate = QtCore.QCoreApplication.translate
        cheatSheet.setWindowTitle(_translate("cheatSheet", "Form"))
        self.txtFilter.setPlaceholderText(_translate("cheatSheet", "Filter (type the name of anything in your project)"))




#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtWidgets import QWidget, QAbstractItemView
from PyQt5.QtCore import QModelIndex

from manuskript.enums import Outline
from manuskript.ui.views.metadataView_ui import Ui_metadataView
from manuskript.ui import style

class metadataView(QWidget, Ui_metadataView):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)
        self._lastIndexes = None
        self.txtSummarySentence.setColumn(Outline.summarySentence)
        self.txtSummaryFull.setColumn(Outline.summaryFull)
        self.txtNotes.setColumn(Outline.notes)
        self.revisions.setEnabled(False)

        self.txtSummarySentence.setStyleSheet(style.lineEditSS())
        self.txtSummaryFull.setStyleSheet(style.transparentSS() +
                                          style.simpleScrollBarV())
        self.txtNotes.setStyleSheet(style.transparentSS() +
                                    style.simpleScrollBarV())
        self.revisions.setStyleSheet(style.simpleScrollBarV())

    def setModels(self, mdlOutline, mdlCharacter, mdlLabels, mdlStatus):
        self.properties.setModels(mdlOutline, mdlCharacter, mdlLabels, mdlStatus)
        self.txtSummarySentence.setModel(mdlOutline)
        self.txtSummaryFull.setModel(mdlOutline)
        self.txtNotes.setModel(mdlOutline)
        self.revisions.setModel(mdlOutline)

    def getIndexes(self, sourceView):
        """Returns a list of indexes from list of QItemSelectionRange"""
        indexes = []

        for i in sourceView.selection().indexes():
            if i.column() != 0:
                continue

            if i not in indexes:
                indexes.append(i)

        return indexes

    def selectionChanged(self):

        if isinstance(self.sender(), QAbstractItemView):
            selectionModel = self.sender().selectionModel()
        else:
            selectionModel = self.sender()

        indexes = self.getIndexes(selectionModel)

        if self._lastIndexes == indexes:
            return

        # No item selected
        if len(indexes) == 0:
            self.setEnabled(False)
            self.revisions.setEnabled(False)
            self.txtSummarySentence.setCurrentModelIndex(QModelIndex())
            self.txtSummaryFull.setCurrentModelIndex(QModelIndex())
            self.txtNotes.setCurrentModelIndex(QModelIndex())

        # One item selected
        elif len(indexes) == 1:
            self.setEnabled(True)
            idx = indexes[0]
            self.txtSummarySentence.setEnabled(True)
            self.txtSummaryFull.setEnabled(True)
            self.txtNotes.setEnabled(True)
            self.txtSummarySentence.setCurrentModelIndex(idx)
            self.txtSummaryFull.setCurrentModelIndex(idx)
            self.txtNotes.setCurrentModelIndex(idx)
            self.revisions.setEnabled(True)
            self.revisions.setCurrentModelIndex(idx)

        # Multiple items selected
        else:
            self.setEnabled(True)

            # Behavior 1:
            # We disable the text areas when multiple indexes are selected
            self.txtSummarySentence.setEnabled(False)
            self.txtSummaryFull.setEnabled(False)
            self.txtNotes.setEnabled(False)
            self.txtSummarySentence.setCurrentModelIndex(QModelIndex())
            self.txtSummaryFull.setCurrentModelIndex(QModelIndex())
            self.txtNotes.setCurrentModelIndex(QModelIndex())

            # Behavior 2:
            # Allow edition of multiple indexes.
            # Bug: Multiple selections of items sometimes gets Notes/references
            #      field to be erased. See #10 on github.
            #self.txtSummarySentence.setCurrentModelIndexes(indexes)
            #self.txtSummaryFull.setCurrentModelIndexes(indexes)
            #self.txtNotes.setCurrentModelIndexes(indexes)

            self.revisions.setEnabled(False)

        self.properties.selectionChanged(selectionModel)
        self._lastIndexes = indexes

    def setDict(self, d):
        self.txtNotes.setDict(d)
        self.txtSummaryFull.setDict(d)

    def toggleSpellcheck(self, v):
        self.txtNotes.toggleSpellcheck(v)
        self.txtSummaryFull.toggleSpellcheck(v)

    def saveState(self):
        return [
            self.grpProperties.saveState(),
            self.grpSummary.saveState(),
            self.grpNotes.saveState(),
            self.grpRevisions.saveState(),
            self.revisions.saveState()
        ]

    def restoreState(self, state):
        self.grpProperties.restoreState(state[0])
        self.grpSummary.restoreState(state[1])
        self.grpNotes.restoreState(state[2])
        self.grpRevisions.restoreState(state[3])



#!/usr/bin/env python
# --!-- coding: utf8 --!--

"""Tests for the welcome widget."""

def test_autoLoad(MWNoProject):
    """
    Tests for the welcome widget using MainWindow with no open project.
    """
    MW = MWNoProject
    from PyQt5.QtCore import QSettings

    # Testing when no autoLoad
    QSettings().remove("autoLoad")
    autoLoad, path = MW.welcome.getAutoLoadValues()
    assert type(autoLoad) == bool
    assert autoLoad == False

    for v in [True, False, 42, "42", None, True]:
        MW.welcome.setAutoLoad(v)
        autoLoad, path = MW.welcome.getAutoLoadValues()
        assert type(autoLoad) == bool



#!/usr/bin/env python
# --!-- coding: utf8 --!--

"""Fixtures."""

import pytest

@pytest.fixture
def MW():
    """
    Returns the mainWindow
    """
    from manuskript import functions as F
    MW = F.mainWindow()
    assert MW is not None
    assert MW == F.MW

    return MW

@pytest.fixture
def MWNoProject(MW):
    """
    Take the MainWindow and close andy possibly open project.
    """
    MW.closeProject()
    assert MW.currentProject is None
    return MW

@pytest.fixture
def MWEmptyProject(MW):
    """
    Creates a MainWindow and load an empty project.
    """
    import tempfile
    tf = tempfile.NamedTemporaryFile(suffix=".msk")

    MW.closeProject()
    assert MW.currentProject is None
    MW.welcome.createFile(tf.name, overwrite=True)
    assert MW.currentProject is not None
    return MW

    # If using with: @pytest.fixture(scope='session', autouse=True)
    # yield MW
    # # Properly destructed after. Otherwise: seg fault.
    # MW.deleteLater()

@pytest.fixture
def MWSampleProject(MW):
    """
    Creates a MainWindow and load a copy of the Acts sample project.
    """

    from manuskript import functions as F
    import os
    # Get the path of the first sample project. We assume it is here.
    spDir = F.appPath("sample-projects")
    lst = os.listdir(spDir)
    # We assume it's saved in folder, so there is a `name.msk` file and a
    # `name` folder.
    src = [f for f in lst if f[-4:] == ".msk" and f[:-4] in lst][0]
    src = os.path.join(spDir, src)
    # Copy to a temp file
    import tempfile
    tf = tempfile.NamedTemporaryFile(suffix=".msk")
    import shutil
    shutil.copyfile(src, tf.name)
    shutil.copytree(src[:-4], tf.name[:-4])
    MW.loadProject(tf.name)
    assert MW.currentProject is not None

    return MW



#!/usr/bin/env python
# --!-- coding: utf8 --!--
import os
import shutil
import subprocess

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import qApp, QMessageBox
from PyQt5.QtGui import QCursor

from manuskript.converters import abstractConverter
from manuskript.functions import mainWindow


class pandocConverter(abstractConverter):

    name = "pandoc"
    cmd = "pandoc"

    @classmethod
    def isValid(self):
        if self.path() != None:
            return 2
        elif self.customPath() and os.path.exists(self.customPath):
            return 1
        else:
            return 0

    @classmethod
    def customPath(self):
        settings = QSettings()
        return settings.value("Exporters/{}_customPath".format(self.name), "")

    @classmethod
    def path(self):
        return shutil.which(self.cmd)

    @classmethod
    def convert(self, src, _from="markdown", to="html", args=None, outputfile=None):
        if not self.isValid:
            print("ERROR: pandocConverter is called but not valid.")
            return ""

        cmd = [self.runCmd()]

        cmd += ["--from={}".format(_from)]
        cmd += ["--to={}".format(to)]

        if args:
            cmd += args

        if outputfile:
            cmd.append("--output={}".format(outputfile))

        qApp.setOverrideCursor(QCursor(Qt.WaitCursor))

        p = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        if not type(src) == bytes:
            src = src.encode("utf-8")  # assumes utf-8

        stdout, stderr = p.communicate(src)

        qApp.restoreOverrideCursor()

        if stderr:
            err = stderr.decode("utf-8")
            print(err)
            QMessageBox.critical(mainWindow().dialog,
                                 qApp.translate("Export", "Error"), err)
            return None

        return stdout.decode("utf-8")

    @classmethod
    def runCmd(self):
        if self.isValid() == 2:
            return self.cmd
        elif self.isValid() == 1:
            return self.customPath



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/welcome_ui.ui'
#
# Created: Mon Oct 16 10:45:41 2017
#      by: PyQt5 UI code generator 5.2.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_welcome(object):
    def setupUi(self, welcome):
        welcome.setObjectName("welcome")
        welcome.resize(728, 459)
        self.horizontalLayout = QtWidgets.QHBoxLayout(welcome)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.frame_2 = QtWidgets.QFrame(welcome)
        self.frame_2.setMinimumSize(QtCore.QSize(500, 0))
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout_32 = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout_32.setObjectName("verticalLayout_32")
        self.horizontalLayout_23 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_23.setObjectName("horizontalLayout_23")
        self.tree = QtWidgets.QTreeWidget(self.frame_2)
        self.tree.setMaximumSize(QtCore.QSize(150, 16777215))
        self.tree.setRootIsDecorated(False)
        self.tree.setObjectName("tree")
        item_0 = QtWidgets.QTreeWidgetItem(self.tree)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        item_0.setFont(0, font)
        brush = QtGui.QBrush(QtGui.QColor(230, 230, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        item_0.setBackground(0, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 128))
        brush.setStyle(QtCore.Qt.NoBrush)
        item_0.setForeground(0, brush)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_0 = QtWidgets.QTreeWidgetItem(self.tree)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        item_0.setFont(0, font)
        brush = QtGui.QBrush(QtGui.QColor(230, 230, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        item_0.setBackground(0, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 128))
        brush.setStyle(QtCore.Qt.NoBrush)
        item_0.setForeground(0, brush)
        self.tree.header().setVisible(False)
        self.horizontalLayout_23.addWidget(self.tree)
        self.templateLayout = QtWidgets.QVBoxLayout()
        self.templateLayout.setObjectName("templateLayout")
        self.lytTemplate = QtWidgets.QGridLayout()
        self.lytTemplate.setObjectName("lytTemplate")
        self.templateLayout.addLayout(self.lytTemplate)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.templateLayout.addItem(spacerItem)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.lblTotal = QtWidgets.QLabel(self.frame_2)
        self.lblTotal.setText("")
        self.lblTotal.setObjectName("lblTotal")
        self.horizontalLayout_2.addWidget(self.lblTotal)
        self.btnAddLevel = QtWidgets.QPushButton(self.frame_2)
        self.btnAddLevel.setEnabled(False)
        icon = QtGui.QIcon.fromTheme("list-add")
        self.btnAddLevel.setIcon(icon)
        self.btnAddLevel.setObjectName("btnAddLevel")
        self.horizontalLayout_2.addWidget(self.btnAddLevel)
        self.btnAddWC = QtWidgets.QPushButton(self.frame_2)
        self.btnAddWC.setEnabled(False)
        icon = QtGui.QIcon.fromTheme("applications-debugging")
        self.btnAddWC.setIcon(icon)
        self.btnAddWC.setObjectName("btnAddWC")
        self.horizontalLayout_2.addWidget(self.btnAddWC)
        self.templateLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_23.addLayout(self.templateLayout)
        self.verticalLayout_32.addLayout(self.horizontalLayout_23)
        self.line_4 = QtWidgets.QFrame(self.frame_2)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.verticalLayout_32.addWidget(self.line_4)
        self.chkLoadLastProject = QtWidgets.QCheckBox(self.frame_2)
        self.chkLoadLastProject.setChecked(True)
        self.chkLoadLastProject.setObjectName("chkLoadLastProject")
        self.verticalLayout_32.addWidget(self.chkLoadLastProject)
        self.horizontalLayout_20 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_20.setObjectName("horizontalLayout_20")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_20.addItem(spacerItem1)
        self.btnOpen = QtWidgets.QPushButton(self.frame_2)
        icon = QtGui.QIcon.fromTheme("document-open")
        self.btnOpen.setIcon(icon)
        self.btnOpen.setObjectName("btnOpen")
        self.horizontalLayout_20.addWidget(self.btnOpen)
        self.btnRecent = QtWidgets.QPushButton(self.frame_2)
        icon = QtGui.QIcon.fromTheme("folder-recent")
        self.btnRecent.setIcon(icon)
        self.btnRecent.setObjectName("btnRecent")
        self.horizontalLayout_20.addWidget(self.btnRecent)
        self.btnCreate = QtWidgets.QPushButton(self.frame_2)
        icon = QtGui.QIcon.fromTheme("dialog-ok")
        self.btnCreate.setIcon(icon)
        self.btnCreate.setObjectName("btnCreate")
        self.horizontalLayout_20.addWidget(self.btnCreate)
        self.verticalLayout_32.addLayout(self.horizontalLayout_20)
        self.horizontalLayout.addWidget(self.frame_2)

        self.retranslateUi(welcome)
        QtCore.QMetaObject.connectSlotsByName(welcome)

    def retranslateUi(self, welcome):
        _translate = QtCore.QCoreApplication.translate
        welcome.setWindowTitle(_translate("welcome", "Form"))
        self.tree.headerItem().setText(0, _translate("welcome", "1"))
        __sortingEnabled = self.tree.isSortingEnabled()
        self.tree.setSortingEnabled(False)
        self.tree.topLevelItem(0).setText(0, _translate("welcome", "Templates"))
        self.tree.topLevelItem(0).child(0).setText(0, _translate("welcome", "Empty"))
        self.tree.topLevelItem(0).child(1).setText(0, _translate("welcome", "Novel"))
        self.tree.topLevelItem(0).child(2).setText(0, _translate("welcome", "Novella"))
        self.tree.topLevelItem(0).child(3).setText(0, _translate("welcome", "Short Story"))
        self.tree.topLevelItem(0).child(4).setText(0, _translate("welcome", "Research paper"))
        self.tree.topLevelItem(1).setText(0, _translate("welcome", "Demo projects"))
        self.tree.setSortingEnabled(__sortingEnabled)
        self.btnAddLevel.setText(_translate("welcome", "Add level"))
        self.btnAddWC.setText(_translate("welcome", "Add word count"))
        self.chkLoadLastProject.setText(_translate("welcome", "Next time, automatically open last project"))
        self.btnOpen.setText(_translate("welcome", "Open..."))
        self.btnRecent.setText(_translate("welcome", "Recent"))
        self.btnCreate.setText(_translate("welcome", "Create"))




#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QLabel
from manuskript.ui.views.webView import webView, webEngine

from manuskript.functions import appPath


if webEngine == "QtWebKit":

    from PyQt5.QtWebKit import QWebSettings
    from PyQt5.QtWebKitWidgets import QWebView


    class PDFViewer(QWebView):
        pdf_viewer_page = "file://"+appPath('libs/pdf.js/web/viewer.html')

        def __init__(self, parent=None):
            QWebView.__init__(self, parent)
            self.settings = QWebSettings.globalSettings()
            self.settings.setAttribute(QWebSettings.LocalContentCanAccessFileUrls, True)

        def loadPDF(self, pdf):
            url = QUrl(self.pdf_viewer_page+"?file="+pdf)
            self.settings.clearMemoryCaches()
            self.load(url)


elif webEngine == "QtWebEngine":

    from PyQt5.QtWebEngineWidgets import QWebEngineView, QWebEnginePage, QWebEngineSettings


    class PDFViewer(QWebEngineView):
        pdf_viewer_page = "file://"+appPath('libs/pdf.js/web/viewer.html')

        def __init__(self, parent=None):
            QWebEngineView.__init__(self, parent)
            self.settings = QWebEngineSettings.globalSettings()
            self.settings.setAttribute(QWebEngineSettings.LocalContentCanAccessFileUrls, True)

        def loadPDF(self, pdf):
            url = QUrl(self.pdf_viewer_page+"?file="+pdf)
            self.load(url)


else:

    class PDFViewer(QLabel):
        def __init__(self, parent=None):
            QLabel.__init__(self, parent)
            self.setText("No Web Engine installed capable of displaying PDF.\n\n"
                         "Consider installing QtWebKit or QtWebEngine.")

        def loadPDF(self, pdf):
            pass



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/editors/tabSplitter_ui.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_tabSplitter(object):
    def setupUi(self, tabSplitter):
        tabSplitter.setObjectName("tabSplitter")
        tabSplitter.resize(400, 300)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(tabSplitter.sizePolicy().hasHeightForWidth())
        tabSplitter.setSizePolicy(sizePolicy)
        tabSplitter.setWindowTitle("tabSPlitter")
        self.verticalLayout = QtWidgets.QVBoxLayout(tabSplitter)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter = QtWidgets.QSplitter(tabSplitter)
        self.splitter.setMinimumSize(QtCore.QSize(30, 30))
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.tab = QtWidgets.QTabWidget(self.splitter)
        self.tab.setTabsClosable(True)
        self.tab.setMovable(True)
        self.tab.setObjectName("tab")
        self.verticalLayout.addWidget(self.splitter)

        self.retranslateUi(tabSplitter)
        QtCore.QMetaObject.connectSlotsByName(tabSplitter)

    def retranslateUi(self, tabSplitter):
        pass




#!/usr/bin/env python
# --!-- coding: utf8 --!--

import locale
import imp
import os

from PyQt5.QtCore import QSettings, QRegExp, Qt, QDir
from PyQt5.QtGui import QIcon, QBrush, QColor, QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QWidget, QAction, QFileDialog, QSpinBox, QLineEdit, QLabel, QPushButton, QTreeWidgetItem, \
    qApp, QMessageBox

from manuskript import loadSave
from manuskript import settings
from manuskript.enums import Outline
from manuskript.functions import mainWindow, iconFromColor, appPath
from manuskript.models.characterModel import characterModel
from manuskript.models import outlineItem, outlineModel
from manuskript.models.plotModel import plotModel
from manuskript.models.worldModel import worldModel
from manuskript.ui.welcome_ui import Ui_welcome
from manuskript.ui import style as S

try:
    locale.setlocale(locale.LC_ALL, '')
except:
    pass

class welcome(QWidget, Ui_welcome):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)

        self.template = []

        self.mw = mainWindow()
        self.btnOpen.clicked.connect(self.openFile)
        self.btnCreate.clicked.connect(self.createFile)
        self.chkLoadLastProject.toggled.connect(self.setAutoLoad)
        self.tree.itemClicked.connect(self.changeTemplate)
        self.btnAddLevel.clicked.connect(self.templateAddLevel)
        self.btnAddWC.clicked.connect(self.templateAddWordCount)
        self.btnCreateText = self.btnCreate.text()

        self.populateTemplates()
        self._templates = self.templates()

    def updateValues(self):
        # Auto load
        autoLoad, last = self.getAutoLoadValues()
        self.chkLoadLastProject.setChecked(autoLoad)

        # Recent Files
        self.loadRecents()

    def getLastAccessedDirectory(self):
        sttgs = QSettings()
        lastDirectory = sttgs.value("lastAccessedDirectory", defaultValue=".", type=str)
        if lastDirectory != '.':
            print(qApp.translate("lastAccessedDirectoryInfo", "Last accessed directory \"{}\" loaded.").format(
                lastDirectory))
        return lastDirectory

    def setLastAccessedDirectory(self, dir):
        sttgs = QSettings()
        sttgs.setValue("lastAccessedDirectory", dir)

    ###############################################################################
    # AUTOLOAD
    ###############################################################################

    def showEvent(self, event):
        """Waiting for things to be fully loaded to start opening projects."""
        QWidget.showEvent(self, event)

        # Auto load last project
        autoLoad, last = self.getAutoLoadValues()

        if self.mw._autoLoadProject:
            project = self.mw._autoLoadProject
            self.mw._autoLoadProject = None
            self.appendToRecentFiles(project)
            self.mw.loadProject(project)

        elif autoLoad and last:
            self.mw.loadProject(last)

    def getAutoLoadValues(self):
        """
        Reads manuskript system's settings and returns a tuple:
        - `bool`: whether manuskript should automatically load
                  the last opened project or display the
                  welcome widget.
        - `str`:  the absolute path to the last opened project.
        """
        sttgns = QSettings()
        autoLoad = sttgns.value("autoLoad", defaultValue=False, type=bool)
        if autoLoad and sttgns.contains("lastProject"):
            last = sttgns.value("lastProject")
        else:
            last = ""

        return autoLoad, last

    def setAutoLoad(self, v):
        if type(v) == bool:
            QSettings().setValue("autoLoad", v)

    ###############################################################################
    # RECENTS
    ###############################################################################

    def loadRecents(self):
        sttgns = QSettings()
        self.mw.menuRecents.setIcon(QIcon.fromTheme("folder-recent"))
        if sttgns.contains("recentFiles"):
            lst = sttgns.value("recentFiles")
            self.mw.menuRecents.clear()
            for f in [f for f in lst if os.path.exists(f)]:
                name = os.path.split(f)[1]
                a = QAction(name, self)
                a.setData(f)
                a.setStatusTip(f)
                a.triggered.connect(self.loadRecentFile)
                self.mw.menuRecents.addAction(a)

            self.btnRecent.setMenu(self.mw.menuRecents)

    def appendToRecentFiles(self, project):
        sttgns = QSettings()
        if sttgns.contains("recentFiles"):
            recentFiles = sttgns.value("recentFiles")
        else:
            recentFiles = []

        while project in recentFiles:
            recentFiles.remove(project)
        recentFiles.insert(0, project)
        recentFiles = recentFiles[:10]
        sttgns.setValue("recentFiles", recentFiles)

    def loadRecentFile(self):
        act = self.sender()
        self.appendToRecentFiles(act.data())
        self.mw.closeProject()
        self.mw.loadProject(act.data())

    ###############################################################################
    # DIALOGS
    ###############################################################################

    def openFile(self):
        lastDirectory = self.getLastAccessedDirectory()

        """File dialog that request an existing file. For opening project."""
        filename = QFileDialog.getOpenFileName(self,
                                               self.tr("Open project"),
                                               lastDirectory,
                                               self.tr("Manuskript project (*.msk);;All files (*)"))[0]
        if filename:
            self.setLastAccessedDirectory(os.path.dirname(filename))
            self.appendToRecentFiles(filename)
            self.mw.loadProject(filename)

    def saveAsFile(self):
        lastDirectory = self.getLastAccessedDirectory()

        """File dialog that request a file, existing or not.
        Save data to that file, which then becomes the current project."""
        filename = QFileDialog.getSaveFileName(self,
                                               self.tr("Save project as..."),
                                               lastDirectory,
                                               self.tr("Manuskript project (*.msk)"))[0]

        if filename:
            self.setLastAccessedDirectory(os.path.dirname(filename))
            if filename[-4:] != ".msk":
                filename += ".msk"
            self.appendToRecentFiles(filename)
            loadSave.clearSaveCache()  # Ensure all file(s) are saved under new filename
            self.mw.saveDatas(filename)
            # Update Window's project name with new filename
            pName = os.path.split(filename)[1]
            if pName.endswith('.msk'):
                pName=pName[:-4]
            self.mw.setWindowTitle(pName + " - " + self.tr("Manuskript"))

    def createFile(self, filename=None, overwrite=False):
        lastDirectory = self.getLastAccessedDirectory()

        """When starting a new project, ask for a place to save it.
        Datas are not loaded from file, so they must be populated another way."""
        if not filename:
            filename = QFileDialog.getSaveFileName(
                           self,
                           self.tr("Create New Project"),
                           lastDirectory,
                           self.tr("Manuskript project (*.msk)"))[0]

        if filename:
            self.setLastAccessedDirectory(os.path.dirname(filename))
            if filename[-4:] != ".msk":
                filename += ".msk"
            if os.path.exists(filename) and not overwrite:
                # Check if okay to overwrite existing project
                result = QMessageBox.warning(self, self.tr("Warning"),
                    self.tr("Overwrite existing project {} ?").format(filename),
                    QMessageBox.Ok|QMessageBox.Cancel, QMessageBox.Cancel)
                if result == QMessageBox.Cancel:
                    return
            # Create new project
            self.appendToRecentFiles(filename)
            self.loadDefaultDatas()
            self.mw.loadProject(filename, loadFromFile=False)

    ###############################################################################
    # TEMPLATES
    ###############################################################################

    def templates(self):
        return [
            (self.tr("Empty fiction"), [], "Fiction"),
            (self.tr("Novel"), [
                (20, self.tr("Chapter")),
                (5, self.tr("Scene")),
                (500, None)  # A line with None is word count
            ], "Fiction"),
            (self.tr("Novella"), [
                (10, self.tr("Chapter")),
                (5, self.tr("Scene")),
                (500, None)
            ], "Fiction"),
            (self.tr("Short Story"), [
                (10, self.tr("Scene")),
                (1000, None)
            ], "Fiction"),
            (self.tr("Trilogy"), [
                (3, self.tr("Book")),
                (3, self.tr("Section")),
                (10, self.tr("Chapter")),
                (5, self.tr("Scene")),
                (500, None)
            ], "Fiction"),
            (self.tr("Empty non-fiction"), [], "Non-fiction"),
            (self.tr("Research paper"), [
                (3, self.tr("Section")),
                (1000, None)
            ], "Non-fiction")
        ]

    def changeTemplate(self, item, column):
        template = [i for i in self._templates if i[0] == item.text(0)]
        self.btnCreate.setText(self.btnCreateText)

        # Selected item is a template
        if len(template):
            self.template = template[0]
            self.updateTemplate()

        # Selected item is a sample project
        elif item.data(0, Qt.UserRole):
            name = item.data(0, Qt.UserRole)
            # Clear templates
            self.template = self._templates[0]
            self.updateTemplate()
            # Change button text
            self.btnCreate.setText("Open {}".format(name))
            # Load project
            self.mw.loadProject(appPath("sample-projects/{}".format(name)))

    def updateTemplate(self):
        # Clear layout
        def clearLayout(l):
            while l.count() != 0:
                i = l.takeAt(0)
                if i.widget():
                    i.widget().deleteLater()
                if i.layout():
                    clearLayout(i.layout())

        clearLayout(self.lytTemplate)

        # self.templateLayout.addStretch()
        # l = QGridLayout()
        # self.templateLayout.addLayout(l)

        k = 0
        hasWC = False
        for templateIndex, d in enumerate(self.template[1]):
            spin = QSpinBox(self)
            spin.setRange(0, 999999)
            spin.setValue(d[0])
            # Storing the level of the template in that spinbox, so we can use
            # it to update the template when valueChanged on that spinbox
            # (we do that in self.updateWordCount for convenience).
            spin.setProperty("templateIndex", templateIndex)
            spin.valueChanged.connect(self.updateWordCount)
            if d[1] != None:
                txt = QLineEdit(self)
                txt.setProperty("templateIndex", templateIndex)
                txt.textEdited.connect(self.updateWordCount)
                txt.setText(d[1])

            else:
                hasWC = True
                txt = QLabel(self.tr("words each."), self)

            if k != 0:
                of = QLabel(self.tr("of"), self)
                self.lytTemplate.addWidget(of, k, 0)

                btn = QPushButton("", self)
                btn.setIcon(QIcon.fromTheme("edit-delete"))
                btn.setProperty("deleteRow", k)
                btn.setFlat(True)
                btn.clicked.connect(self.deleteTemplateRow)

                self.lytTemplate.addWidget(btn, k, 3)

            self.lytTemplate.addWidget(spin, k, 1)
            self.lytTemplate.addWidget(txt, k, 2)
            k += 1

        self.btnAddWC.setEnabled(not hasWC and len(self.template[1]) > 0)
        self.btnAddLevel.setEnabled(True)
        self.lblTotal.setVisible(hasWC)
        self.updateWordCount()

    def templateAddLevel(self):
        if len(self.template[1]) > 0 and \
                        self.template[1][len(self.template[1]) - 1][1] == None:
            # has word count, so insert before
            self.template[1].insert(len(self.template[1]) - 1, (10, self.tr("Text")))
        else:
            # No word count, so insert at end
            self.template[1].append((10, self.tr("Something")))
        self.updateTemplate()

    def templateAddWordCount(self):
        self.template[1].append((500, None))
        self.updateTemplate()

    def deleteTemplateRow(self):
        btn = self.sender()
        row = btn.property("deleteRow")
        self.template[1].pop(row)
        self.updateTemplate()

    def updateWordCount(self):
        """
        Updates the word count of the template, and displays it in a label.

        Also, updates self.template, which is used to create the items when
        calling self.createFile.
        """
        total = 1

        # Searching for every spinboxes on the widget, and multiplying
        # their values to get the number of words.
        for s in self.findChildren(QSpinBox, QRegExp(".*"),
                                   Qt.FindChildrenRecursively):
            total = total * s.value()

            # Update self.template to reflect the changed count values
            templateIndex = s.property("templateIndex")
            self.template[1][templateIndex] = (
                s.value(),
                self.template[1][templateIndex][1])

        for t in self.findChildren(QLineEdit, QRegExp(".*"),
                                   Qt.FindChildrenRecursively):
            # Update self.template to reflect the changed name values
            templateIndex = t.property("templateIndex")
            if templateIndex is not None :
                self.template[1][templateIndex] = (
                self.template[1][templateIndex][0],
                t.text())

        if total == 1:
            total = 0

        self.lblTotal.setText(self.tr("<b>Total:</b> {} words (~ {} pages)").format(
                locale.format_string("%d", total, grouping=True),
                locale.format_string("%d", total / 250, grouping=True)
        ))

    def addTopLevelItem(self, name):
        item = QTreeWidgetItem(self.tree, [name])
        item.setBackground(0, QBrush(QColor(S.highlightLight)))
        item.setForeground(0, QBrush(QColor(S.highlightedTextDark)))
        item.setTextAlignment(0, Qt.AlignCenter)
        item.setFlags(Qt.ItemIsEnabled)
        f = item.font(0)
        f.setBold(True)
        item.setFont(0, f)
        return item

    def populateTemplates(self):
        self.tree.clear()
        self.tree.setIndentation(0)

        # Add templates
        item = self.addTopLevelItem(self.tr("Fiction"))
        templates = [i for i in self.templates() if i[2] == "Fiction"]
        for t in templates:
            sub = QTreeWidgetItem(item, [t[0]])

        # Add templates: non-fiction
        item = self.addTopLevelItem(self.tr("Non-fiction"))
        templates = [i for i in self.templates() if i[2] == "Non-fiction"]
        for t in templates:
            sub = QTreeWidgetItem(item, [t[0]])


        # Add Demo project
        item = self.addTopLevelItem(self.tr("Demo projects"))
        dir = QDir(appPath("sample-projects"))
        for f in dir.entryList(["*.msk"], filters=QDir.Files):
            sub = QTreeWidgetItem(item, [f[:-4]])
            sub.setData(0, Qt.UserRole, f)

        self.tree.expandAll()

    def loadDefaultDatas(self):

        # Empty settings
        imp.reload(settings)
        settings.initDefaultValues()

        if self.template:
            t = [i for i in self._templates if i[0] == self.template[0]]
            if t and t[0][2] == "Non-fiction":
                settings.viewMode = "simple"

        # Tasks
        self.mw.mdlFlatData = QStandardItemModel(2, 8, self.mw)

        # Persos
        # self.mw.mdlPersos = QStandardItemModel(0, 0, self.mw)
        self.mw.mdlCharacter = characterModel(self.mw)
        # self.mdlPersosProxy = None # persosProxyModel() # None
        # self.mw.mdlPersosProxy = persosProxyModel(self.mw)

        # self.mw.mdlPersosInfos = QStandardItemModel(1, 0, self.mw)
        # self.mw.mdlPersosInfos.insertColumn(0, [QStandardItem("ID")])
        # self.mw.mdlPersosInfos.setHorizontalHeaderLabels(["Description"])

        # Labels
        self.mw.mdlLabels = QStandardItemModel(self.mw)
        for color, text in [
            (Qt.transparent, ""),
            (Qt.yellow, self.tr("Idea")),
            (Qt.green, self.tr("Note")),
            (Qt.blue, self.tr("Chapter")),
            (Qt.red, self.tr("Scene")),
            (Qt.cyan, self.tr("Research"))
        ]:
            self.mw.mdlLabels.appendRow(QStandardItem(iconFromColor(color), text))

        # Status
        self.mw.mdlStatus = QStandardItemModel(self.mw)
        for text in [
            "",
            self.tr("TODO"),
            self.tr("First draft"),
            self.tr("Second draft"),
            self.tr("Final")
        ]:
            self.mw.mdlStatus.appendRow(QStandardItem(text))

        # Plot
        self.mw.mdlPlots = plotModel(self.mw)

        # Outline
        self.mw.mdlOutline = outlineModel(self.mw)

        # World
        self.mw.mdlWorld = worldModel(self.mw)

        root = self.mw.mdlOutline.rootItem
        _type = "md"

        def addElement(parent, datas):
            if len(datas) == 2 and datas[1][1] == None or \
                            len(datas) == 1:
                # Next item is word count
                n = 0
                for i in range(datas[0][0]):
                    n += 1
                    item = outlineItem(title="{} {}".format(
                            datas[0][1],
                            str(n)),
                            _type=_type,
                            parent=parent)
                    if len(datas) == 2:
                        item.setData(Outline.setGoal, datas[1][0])
                        # parent.appendChild(item)
            else:
                n = 0
                for i in range(datas[0][0]):
                    n += 1
                    item = outlineItem(title="{} {}".format(
                            datas[0][1],
                            str(n)),
                            _type="folder",
                            parent=parent)
                    # parent.appendChild(item)
                    addElement(item, datas[1:])

        if self.template and self.template[1]:
            addElement(root, self.template[1])



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import Qt, QSize, QModelIndex
from PyQt5.QtGui import QColor, QPalette, QIcon, QFont, QFontMetrics, QBrush
from PyQt5.QtWidgets import QStyledItemDelegate, QStyleOptionViewItem, QStyle, QComboBox, QStyleOptionComboBox
from PyQt5.QtWidgets import qApp

from manuskript import settings
from manuskript.enums import Character, Outline
from manuskript.functions import outlineItemColors, mixColors, colorifyPixmap, toInt, toFloat, drawProgress
from manuskript.ui import style as S


class outlineTitleDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)
        self._view = None

    def setView(self, view):
        self._view = view

    def paint(self, painter, option, index):

        item = index.internalPointer()
        colors = outlineItemColors(item)

        style = qApp.style()

        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)

        iconRect = style.subElementRect(style.SE_ItemViewItemDecoration, opt)
        textRect = style.subElementRect(style.SE_ItemViewItemText, opt)

        # Background
        style.drawPrimitive(style.PE_PanelItemViewItem, opt, painter)

        if settings.viewSettings["Outline"]["Background"] != "Nothing" and not opt.state & QStyle.State_Selected:

            col = colors[settings.viewSettings["Outline"]["Background"]]

            if col != QColor(Qt.transparent):
                col2 = QColor(S.base)
                if opt.state & QStyle.State_Selected:
                    col2 = opt.palette.brush(QPalette.Normal, QPalette.Highlight).color()
                col = mixColors(col, col2, .2)

            painter.save()
            painter.setBrush(col)
            painter.setPen(Qt.NoPen)

            rect = opt.rect
            if self._view:
                r2 = self._view.visualRect(index)
                rect = self._view.viewport().rect()
                rect.setLeft(r2.left())
                rect.setTop(r2.top())
                rect.setBottom(r2.bottom())

            painter.drawRoundedRect(rect, 5, 5)
            painter.restore()

        # Icon
        mode = QIcon.Normal
        if not opt.state & QStyle.State_Enabled:
            mode = QIcon.Disabled
        elif opt.state & QStyle.State_Selected:
            mode = QIcon.Selected
        state = QIcon.On if opt.state & QStyle.State_Open else QIcon.Off
        icon = opt.icon.pixmap(iconRect.size(), mode=mode, state=state)
        if opt.icon and settings.viewSettings["Outline"]["Icon"] != "Nothing":
            color = colors[settings.viewSettings["Outline"]["Icon"]]
            colorifyPixmap(icon, color)
        opt.icon = QIcon(icon)
        opt.icon.paint(painter, iconRect, opt.decorationAlignment, mode, state)

        # Text
        if opt.text:
            painter.save()
            textColor = QColor(S.text)
            if option.state & QStyle.State_Selected:
                col = QColor(S.highlightedText)
                textColor = col
                painter.setPen(col)
            if settings.viewSettings["Outline"]["Text"] != "Nothing":
                col = colors[settings.viewSettings["Outline"]["Text"]]
                if col == Qt.transparent:
                    col = textColor
                # If text color is Compile and item is selected, we have
                # to change the color
                if settings.viewSettings["Outline"]["Text"] == "Compile" and \
                   item.compile() in [0, "0"]:
                    col = mixColors(textColor, QColor(S.window))
                painter.setPen(col)
            f = QFont(opt.font)
            painter.setFont(f)
            fm = QFontMetrics(f)
            elidedText = fm.elidedText(opt.text, Qt.ElideRight, textRect.width())
            painter.drawText(textRect, Qt.AlignLeft, elidedText)

            painter.restore()

            # QStyledItemDelegate.paint(self, painter, option, index)


class outlineCharacterDelegate(QStyledItemDelegate):
    def __init__(self, mdlCharacter, parent=None):
        QStyledItemDelegate.__init__(self, parent)
        self.mdlCharacter = mdlCharacter

    def sizeHint(self, option, index):
        # s = QStyledItemDelegate.sizeHint(self, option, index)

        item = QModelIndex()
        character = self.mdlCharacter.getCharacterByID(index.data())
        if character:
            item = character.index(Character.name)

        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, item)
        s = QStyledItemDelegate.sizeHint(self, opt, item)

        if s.width() > 200:
            s.setWidth(200)
        elif s.width() < 100:
            s.setWidth(100)
        return s + QSize(18, 0)

    def createEditor(self, parent, option, index):
        item = index.internalPointer()
        # if item.isFolder():  # No POV for folders
        # return

        editor = QComboBox(parent)
        editor.setAutoFillBackground(True)
        editor.setFrame(False)
        return editor

    def setEditorData(self, editor, index):
        # editor.addItem("")
        editor.addItem(QIcon.fromTheme("dialog-no"), self.tr("None"))

        l = [self.tr("Main"), self.tr("Secondary"), self.tr("Minor")]
        for importance in range(3):
            editor.addItem(l[importance])
            editor.setItemData(editor.count() - 1, QBrush(QColor(S.highlightedTextDark)), Qt.ForegroundRole)
            editor.setItemData(editor.count() - 1, QBrush(QColor(S.highlightLight)), Qt.BackgroundRole)
            item = editor.model().item(editor.count() - 1)
            item.setFlags(Qt.ItemIsEnabled)
            for i in range(self.mdlCharacter.rowCount()):
                imp = toInt(self.mdlCharacter.importance(i))
                if not 2 - imp == importance: continue

                # try:
                editor.addItem(self.mdlCharacter.icon(i), self.mdlCharacter.name(i), self.mdlCharacter.ID(i))
                editor.setItemData(editor.count() - 1, self.mdlCharacter.name(i), Qt.ToolTipRole)
                # except:
                # pass

        editor.setCurrentIndex(editor.findData(index.data()))
        editor.showPopup()

    def setModelData(self, editor, model, index):
        val = editor.currentData()
        model.setData(index, val)

    def paint(self, painter, option, index):
        ##option.rect.setWidth(option.rect.width() - 18)
        # QStyledItemDelegate.paint(self, painter, option, index)
        ##option.rect.setWidth(option.rect.width() + 18)

        itemIndex = QModelIndex()
        character = self.mdlCharacter.getCharacterByID(index.data())
        if character:
            itemIndex = character.index(Character.name)

        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, itemIndex)

        qApp.style().drawControl(QStyle.CE_ItemViewItem, opt, painter)

        # if index.isValid() and index.internalPointer().data(Outline.POV) not in ["", None]:
        if itemIndex.isValid() and self.mdlCharacter.data(itemIndex) not in ["", None]:
            opt = QStyleOptionComboBox()
            opt.rect = option.rect
            r = qApp.style().subControlRect(QStyle.CC_ComboBox, opt, QStyle.SC_ComboBoxArrow)
            option.rect = r
            qApp.style().drawPrimitive(QStyle.PE_IndicatorArrowDown, option, painter)


class outlineCompileDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def displayText(self, value, locale):
        return ""

    #def createEditor(self, parent, option, index):
        #return None


class outlineGoalPercentageDelegate(QStyledItemDelegate):
    def __init__(self, rootIndex=None, parent=None):
        QStyledItemDelegate.__init__(self, parent)
        self.rootIndex = rootIndex

    def sizeHint(self, option, index):
        sh = QStyledItemDelegate.sizeHint(self, option, index)
        # if sh.width() > 50:
        sh.setWidth(100)
        return sh

    def paint(self, painter, option, index):
        if not index.isValid():
            return QStyledItemDelegate.paint(self, painter, option, index)

        QStyledItemDelegate.paint(self, painter, option, index)

        item = index.internalPointer()

        if not item.data(Outline.goal):
            return

        p = toFloat(item.data(Outline.goalPercentage))

        typ = item.data(Outline.type)

        level = item.level()
        if self.rootIndex and self.rootIndex.isValid():
            level -= self.rootIndex.internalPointer().level() + 1

        margin = 5
        height = max(min(option.rect.height() - 2 * margin, 12) - 2 * level, 6)

        painter.save()

        rect = option.rect.adjusted(margin, margin, -margin, -margin)

        # Move
        rect.translate(level * rect.width() / 10, 0)
        rect.setWidth(rect.width() - level * rect.width() / 10)

        rect.setHeight(height)
        rect.setTop(option.rect.top() + (option.rect.height() - height) / 2)

        drawProgress(painter, rect, p)  # from functions

        painter.restore()

    def displayText(self, value, locale):
        return ""


class outlineStatusDelegate(QStyledItemDelegate):
    def __init__(self, mdlStatus, parent=None):
        QStyledItemDelegate.__init__(self, parent)
        self.mdlStatus = mdlStatus

    def sizeHint(self, option, index):
        s = QStyledItemDelegate.sizeHint(self, option, index)
        if s.width() > 150:
            s.setWidth(150)
        elif s.width() < 50:
            s.setWidth(50)
        return s + QSize(18, 0)

    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        editor.setAutoFillBackground(True)
        editor.setFrame(False)
        return editor

    def setEditorData(self, editor, index):
        for i in range(self.mdlStatus.rowCount()):
            editor.addItem(self.mdlStatus.item(i, 0).text())

        val = index.internalPointer().data(Outline.status)
        if not val: val = 0
        editor.setCurrentIndex(int(val))
        editor.showPopup()

    def setModelData(self, editor, model, index):
        val = editor.currentIndex()
        model.setData(index, val)

    def displayText(self, value, locale):
        try:
            return self.mdlStatus.item(int(value), 0).text()
        except:
            return ""

    def paint(self, painter, option, index):
        QStyledItemDelegate.paint(self, painter, option, index)

        if index.isValid() and index.internalPointer().data(Outline.status) not in ["", None, "0", 0]:
            opt = QStyleOptionComboBox()
            opt.rect = option.rect
            r = qApp.style().subControlRect(QStyle.CC_ComboBox, opt, QStyle.SC_ComboBoxArrow)
            option.rect = r
            qApp.style().drawPrimitive(QStyle.PE_IndicatorArrowDown, option, painter)


class outlineLabelDelegate(QStyledItemDelegate):
    def __init__(self, mdlLabels, parent=None):
        QStyledItemDelegate.__init__(self, parent)
        self.mdlLabels = mdlLabels

    def sizeHint(self, option, index):
        d = index.internalPointer().data(index.column(), Qt.DisplayRole)
        if not d:
            d = 0
        item = self.mdlLabels.item(int(d), 0)
        idx = self.mdlLabels.indexFromItem(item)
        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, idx)
        s = qApp.style().sizeFromContents(QStyle.CT_ItemViewItem, opt, QSize())
        if s.width() > 150:
            s.setWidth(150)
        elif s.width() < 50:
            s.setWidth(50)
        return s + QSize(18, 0)

    def createEditor(self, parent, option, index):
        item = index.internalPointer()
        editor = QComboBox(parent)
        # editor.setAutoFillBackground(True)
        editor.setFrame(False)
        return editor

    def setEditorData(self, editor, index):
        for i in range(self.mdlLabels.rowCount()):
            editor.addItem(self.mdlLabels.item(i, 0).icon(),
                           self.mdlLabels.item(i, 0).text())

        val = index.internalPointer().data(Outline.label)
        if not val: val = 0
        editor.setCurrentIndex(int(val))
        editor.showPopup()

    def setModelData(self, editor, model, index):
        val = editor.currentIndex()
        model.setData(index, val)

    def paint(self, painter, option, index):
        if not index.isValid():
            return QStyledItemDelegate.paint(self, painter, option, index)
        else:
            item = index.internalPointer()

        d = item.data(index.column(), Qt.DisplayRole)
        if not d:
            d = 0

        lbl = self.mdlLabels.item(int(d), 0)
        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, self.mdlLabels.indexFromItem(lbl))

        qApp.style().drawControl(QStyle.CE_ItemViewItem, opt, painter)

        # Drop down indicator
        if index.isValid() and index.internalPointer().data(Outline.label) not in ["", None, "0", 0]:
            opt = QStyleOptionComboBox()
            opt.rect = option.rect
            r = qApp.style().subControlRect(QStyle.CC_ComboBox, opt, QStyle.SC_ComboBoxArrow)
            option.rect = r
            qApp.style().drawPrimitive(QStyle.PE_IndicatorArrowDown, option, painter)



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import pyqtSignal, Qt, QTimer, QRect
from PyQt5.QtGui import QBrush, QCursor, QPalette, QFontMetrics, QColor
from PyQt5.QtWidgets import QWidget, QListWidgetItem, QToolTip, QStyledItemDelegate, QStyle

from manuskript.enums import Character
from manuskript.enums import Plot
from manuskript.functions import mainWindow
from manuskript.ui import style as S
from manuskript.ui.cheatSheet_ui import Ui_cheatSheet
from manuskript.models import references as Ref
from manuskript.ui.editors.completer import completer


class cheatSheet(QWidget, Ui_cheatSheet):
    activated = pyqtSignal(str)

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)
        self.txtFilter.setStyleSheet(S.lineEditSS())
        self.splitter.setStretchFactor(0, 5)
        self.splitter.setStretchFactor(1, 70)

        self.txtFilter.textChanged.connect(self.textChanged)
        self.txtFilter.textChanged.connect(self.updateListFromData)
        self.txtFilter.returnPressed.connect(self.showInfos)
        self.listDelegate = listCompleterDelegate(self)
        self.list.setItemDelegate(self.listDelegate)
        self.list.itemActivated.connect(self.showInfos)
        self.list.itemClicked.connect(self.showInfos)
        self.hideList()
        self.list.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.view.linkActivated.connect(self.openLink)
        self.view.linkHovered.connect(self.linkHovered)
        self.btnShowList.toggled.connect(self.list.setVisible)
        self.line.hide()

        self.outlineModel = None
        self.characterModel = None
        self.plotModel = None
        self.worldModel = None

        self.populateTimer = QTimer(self)
        self.populateTimer.setSingleShot(True)
        self.populateTimer.setInterval(500)
        self.populateTimer.timeout.connect(self.populate)
        self.populateTimer.stop()

        self.data = {}

        self.populate()

    def setModels(self):
        mw = mainWindow()
        self.outlineModel = mw.mdlOutline
        self.characterModel = mw.mdlCharacter
        self.plotModel = mw.mdlPlots
        self.worldModel = mw.mdlWorld

        self.outlineModel.dataChanged.connect(self.populateTimer.start)
        self.characterModel.dataChanged.connect(self.populateTimer.start)
        self.characterModel.rowsInserted.connect(self.populateTimer.start)
        self.characterModel.rowsRemoved.connect(self.populateTimer.start)
        self.plotModel.dataChanged.connect(self.populateTimer.start)
        self.worldModel.dataChanged.connect(self.populateTimer.start)

        self.populate()

    def textChanged(self, text):
        if not text:
            self.hideList()
            self.list.clear()
            self.view.setText("")
        else:
            self.list.show()

    def hideList(self):
        if not self.btnShowList.isChecked():
            self.list.hide()

    def populate(self):
        if self.characterModel:
            d = []

            for c in self.characterModel.characters:
                imp = [self.tr("Minor"), self.tr("Secondary"), self.tr("Main")][int(c.importance())]
                d.append((c.name(), c.ID(), imp))

            self.data[(self.tr("Characters"), Ref.CharacterLetter)] = d

        if self.outlineModel:
            d = []

            def addChildren(item):
                for c in item.children():
                    d.append((c.title(), c.ID(), c.path()))
                    addChildren(c)

            r = self.outlineModel.rootItem
            addChildren(r)

            self.data[(self.tr("Texts"), Ref.TextLetter)] = d

        if self.plotModel:
            d = []

            for r in range(self.plotModel.rowCount()):
                name = self.plotModel.item(r, Plot.name).text()
                ID = self.plotModel.item(r, Plot.ID).text()
                imp = self.plotModel.item(r, Plot.importance).text()
                imp = [self.tr("Minor"), self.tr("Secondary"), self.tr("Main")][int(imp)]
                d.append((name, ID, imp))

            self.data[(self.tr("Plots"), Ref.PlotLetter)] = d

        if self.worldModel:
            d = self.worldModel.listAll()
            self.data[(self.tr("World"), Ref.WorldLetter)] = d

        self.updateListFromData()

    def addCategory(self, title):
        item = QListWidgetItem(title)
        item.setBackground(QBrush(QColor(S.highlightLight)))
        item.setForeground(QBrush(QColor(S.highlightedTextDark)))
        item.setFlags(Qt.ItemIsEnabled)
        f = item.font()
        f.setBold(True)
        item.setFont(f)
        self.list.addItem(item)

    def updateListFromData(self):
        self.list.clear()
        for cat in self.data:
            filtered = [i for i in self.data[cat] if self.txtFilter.text().lower() in i[0].lower() or
                                                     self.txtFilter.text().lower() in cat[0].lower()]
            if filtered:
                self.addCategory(cat[0])
                for item in filtered:
                    i = QListWidgetItem(item[0])
                    i.setData(Qt.UserRole, Ref.EmptyRef.format(cat[1], item[1], item[0]))
                    i.setData(Qt.UserRole + 1, item[2])
                    self.list.addItem(i)

        self.list.setCurrentRow(1)

    def showInfos(self):
        self.hideList()
        if self.list and len(self.txtFilter.text()) != 0:
            i = self.list.currentItem()
            ref = i.data(Qt.UserRole)
            if ref:
                self.view.setText(Ref.infos(ref))

    def openLink(self, link):
        Ref.open(link)

    def linkHovered(self, link):
        if link:
            QToolTip.showText(QCursor.pos(), Ref.tooltip(link))

    def keyPressEvent(self, event):
        if event.key() in [Qt.Key_Up, Qt.Key_Down]:
            self.list.keyPressEvent(event)
        else:
            QWidget.keyPressEvent(self, event)


class listCompleterDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        extra = index.data(Qt.UserRole + 1)
        if not extra:
            return QStyledItemDelegate.paint(self, painter, option, index)

        else:
            if option.state & QStyle.State_Selected:
                painter.fillRect(option.rect, option.palette.color(QPalette.Inactive, QPalette.Highlight))

            title = index.data()
            extra = " - {}".format(extra)
            painter.drawText(option.rect, Qt.AlignLeft, title)

            fm = QFontMetrics(option.font)
            w = fm.width(title)
            r = QRect(option.rect)
            r.setLeft(r.left() + w)
            painter.save()
            painter.setPen(Qt.gray)
            painter.drawText(r, Qt.AlignLeft, extra)
            painter.restore()



#!/usr/bin/env python
# --!-- coding: utf8 --!--


# default window color (linux):
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QPalette
from PyQt5.QtWidgets import qApp

from manuskript import settings
from manuskript import functions as F

# Loading palette colors.
# Manuskript as to restart to reload
p = qApp.palette()
# window = "#d6d2d0" #"#eee" / #eff0f1
window = p.color(QPalette.Window).name()            # General background
windowText = p.color(QPalette.WindowText).name()    # General foreground
base = p.color(QPalette.Base).name()                # Other background
alternateBase = p.color(QPalette.AlternateBase).name() # Other background
text = p.color(QPalette.Text).name()                # Base Text
brightText = p.color(QPalette.BrightText).name()    # Contrast Text
button = p.color(QPalette.Button).name()            # Button background
buttonText = p.color(QPalette.ButtonText).name()    # Button Text
highlight = p.color(QPalette.Highlight).name()      # Other background
highlightedText = p.color(QPalette.HighlightedText).name() # Base Text
link = p.color(QPalette.Link).name()                # Link
linkVisited = p.color(QPalette.LinkVisited).name()  # Link visited

light = p.color(QPalette.Light).name()       # Lighter than Button color
midlight = p.color(QPalette.Midlight).name() # Between Button and Light
mid = p.color(QPalette.Mid).name()           # Between Button and Dark
dark = p.color(QPalette.Dark).name()         # Darker than Button
shadow = p.color(QPalette.Shadow).name()     # A very dark color

highlightLight = F.mixColors(highlight, window, .3)
highlightedTextDark = F.mixColors(highlight, text, .3)
highlightedTextLight = F.mixColors(highlight, highlightedText)
midlighter = F.mixColors(mid, window, .4)
textLight = F.mixColors(window, text)
textLighter = F.mixColors(window, text, .85)


#from manuskript.ui import style as S
#QColor(S.highlightedTextDark)
#QColor(S.highlightLight)


def mainWindowSS():
    return """
    QMenuBar#menubar{{border:none;}}

    QToolButton{{
        background: none;
        border: none;
    }}
    QPushButton:flat:hover, QToolButton:hover{{
        border: 1px solid {borderHover};
        border-radius: 3px;
        background: {backgroundHover};
    }}
    """.format(
        backgroundHover=highlightLight,
        borderHover=mid
    )

def styleMainWindow(mw):
    mw.setStyleSheet(mainWindowSS())
    mw.lstTabs.verticalScrollBar().setStyleSheet(simpleScrollBarV())

    # Custom palette?
    #qApp.setPalette(appPalette())

    mw.treeRedacOutline.setStyleSheet("""
            QTreeView{
                background: transparent;
                margin-top: 30px;
            }""" + simpleScrollBarV())

    mw.lstTabs.setStyleSheet("""
        QListView {{
            show-decoration-selected: 0;
            outline: none;
            background-color: transparent;
        }}

        QListView::item:selected {{
            background: {highlight};
            color: {textSelected}
        }}

        QListView::item:hover {{
            background: {hover};
        }}
        """.format(
            hover=highlight,
            highlight=highlightLight,
            textSelected=text,
        ))


def appPalette():
    p = qApp.palette()
    c = p.color(p.Window)
    # p.setColor(p.Window, QColor(window))
    # p.setColor(p.Base, c.lighter(115))
    # p.setColor(p.Base, QColor("#FFF"))
    return p

    # dark_palette = QPalette()
    # dark_palette.setColor(QPalette.Window, QColor(53, 53, 53))
    # dark_palette.setColor(QPalette.WindowText, Qt.white)
    # dark_palette.setColor(QPalette.Base, QColor(25, 25, 25))
    # dark_palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
    # dark_palette.setColor(QPalette.ToolTipBase, Qt.white)
    # dark_palette.setColor(QPalette.ToolTipText, Qt.white)
    # dark_palette.setColor(QPalette.Text, Qt.white)
    # dark_palette.setColor(QPalette.Button, QColor(53, 53, 53))
    # dark_palette.setColor(QPalette.ButtonText, Qt.white)
    # dark_palette.setColor(QPalette.BrightText, Qt.red)
    # dark_palette.setColor(QPalette.Link, QColor(42, 130, 218))
    # dark_palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    # dark_palette.setColor(QPalette.HighlightedText, Qt.black)
    # qApp.setStyleSheet("QToolTip { color: #ffffff; background-color: #2a82da; border: 1px solid white; }")
    #
    # return dark_palette


def collapsibleGroupBoxButton():
    s1 = """
        QPushButton{{
            background-color: #BBB;
            border: none;
            padding: 2px;
        }}
        QPushButton:checked, QPushButton:hover{{
            font-style:italic;
            background-color:{bg};
        }}""".format(bg=highlightLight)

    s2 = """
        QPushButton{{
            background-color: transparent;
            border: none;
            border-top: 1px solid {border};
            padding: 4px 0px;
            font-weight: bold;
        }}
        QPushButton:hover{{
            background-color:{hover};
        }}
        """.format(
        hover=highlightLight,
        border=mid,
    )

    return s2


def mainEditorTabSS():
    if not settings.textEditor["backgroundTransparent"]:
        SS = """
            QTabWidget::pane{{
                margin-top: -{bw}px;
                border: {bw}px solid {borderColor};
            }}
            QTabWidget::tab-bar{{
                left:50px;
            }}
            QTabBar{{
                background: {bgColor};
                border-radius: 0;
                border: 0px;
            }}
            QTabBar::tab{{
                padding: 2px 9px;
                border: {bw}px solid {borderColor};
                border-bottom: 0px;
            }}
            QTabBar::tab:selected{{
                border: {bw}px solid {borderColor};
                background: {bgColorSelected};
                border-bottom: 0px;
                color: {foreground};
            }}
            QTabBar::tab:!selected:hover{{
                background:{highlight};
                color: {highlightedText};
            }}
            """.format(
                bgColor=textLighter,
                bgColorSelected=settings.textEditor["background"],
                foreground=settings.textEditor["fontColor"],
                borderColor=mid,
                highlight=highlight,
                highlightedText=highlightedText,
                bw=0,
            )
    else:
        # Transparent text view
        SS = """
            QTabWidget::pane{{
                margin-top: -{bw}px;
                border: none;
            }}
            QTabWidget::tab-bar{{
                left:50px;
            }}
            QTabBar{{
                background: {bgColor};
                border: 0px;
            }}
            QTabBar::tab{{
                padding: 2px 9px;
                border: {bw}px solid {borderColor};
            }}
            QTabBar::tab:selected{{
                border: {bw}px solid {borderColor};
                background: {highlight};
                color: {highlightedText};
            }}
            QTabBar::tab:!selected:hover{{
                background:{highlight};
                color: {highlightedText};
            }}
            """.format(
                bgColor=textLighter,
                highlight=highlightLight,
                highlightedText=text,
                text=text,
                borderColor=mid,
                bw=0,
            )

    # Add scrollbar
    SS += simpleScrollBarV(handle=mid, width=10)

    return SS


def toolBarSS():
    return """
        QToolBar{{
            background:transparent;
            border:none;
            spacing: 0px;
            margin-top:40px;
        }}
        QToolBar:separator{{
            border: none;
        }}
        """.format(
            border=midlighter,
            )

def verticalToolButtonSS():
    return """
        QToolButton{{
            border: 0;
            border-radius: 0px;
            background: transparent;
            margin: 0px;
            padding: 4px 8px;
        }}
        QToolButton:checked{{
            border: 0px solid {borderColor};
            background: {bgChecked};
        }}
        QToolButton:hover{{
            border: 0px solid {borderColor};
            background: {bgHover};
        }}
        """.format(
        borderColor=mid,
        bgChecked=textLighter,
        bgHover=highlightLight,
    )


def dockSS():

    return """
        QDockWidget::title {{
            text-align: left; /* align the text to the left */
            background: {header};
            padding: 5px;
        }}

        QDockWidget::close-button, QDockWidget::float-button {{
            background: {header};
        }}
        """.format(
        header=highlightLight,
        button=button
    )


def searchResultSS():
    return """
        QListWidget{{
            background: {window};
        }}
        """.format(
        window=window
    )


def lineEditSS():
    # return "border-radius: 6px;"
    return """QLineEdit{{
        border: none;
        border-bottom: 1px solid {line};
        background:{window};
    }}
    QLineEdit:focus{{
        border-bottom: 1px solid {highlight};
    }}
    """.format(window=window,
               line=mid,
               highlight=highlight)


def transparentSS():
    return """
        QTextEdit{
            background: transparent;
            border:none;
        }"""

def simpleScrollBarV(handle=None, width=8):
    # system default is (i think): mid background, dark handle

    default = midlighter

    handle = handle or default
    return """
        QScrollBar:vertical {{
            border: none;
            background: {background};
            width: {width}px;
        }}
        QScrollBar::handle {{
            background: {handle};
        }}
        QScrollBar::add-line:vertical {{
            width:0;
            height: 0;
            border: none;
            background: none;
        }}

        QScrollBar::sub-line:vertical {{
            width:0;
            height: 0;
            border: none;
            background: none;
        }}""".format(
            background="transparent",
            handle=handle,
            width=width)

def toolBoxSS():
    return """
    QToolBox::tab{{
        background-color: {background};
        padding: 2px;
        border: none;
    }}

    QToolBox::tab:selected, QToolBox::tab:hover{{
        background-color:{backgroundHover};
        color: {colorHover};
    }}""".format(
            background=highlightLight,
            backgroundHover=highlight,
            colorHover=highlightedText,
        )

def titleLabelSS():
    return """
        QLabel{{
            background-color:{bg};
            border:none;
            padding:10px;
            color:{text};
            font-size:16px;
            font-weight:bold;
            text-align:center;
        }}""".format(
            bg=highlightLight,
            text=highlightedTextDark,
            )



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import pyqtSignal, pyqtProperty
from PyQt5.QtWidgets import QWidget
from manuskript.ui.views.sldImportance_ui import Ui_sldImportance

from manuskript.functions import toInt


class sldImportance(QWidget, Ui_sldImportance):
    importanceChanged = pyqtSignal(str)

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)
        self._column = 0
        self._updating = False
        self._index = None

        self.lastValue = -1
        self.sld.valueChanged.connect(self.changed)
        self.setValue(0)

    def getImportance(self):
        return str(self.sld.value())

    def changed(self, v):
        val = [
            self.tr("Minor"),
            self.tr("Secondary"),
            self.tr("Main"),
        ]
        self.lbl.setText(val[v])

        self.importanceChanged.emit(str(v))

        if self._index and not self._updating:
            if str(v) != self._model.data(self._index):
                self._updating = True
                self._model.setData(self._index, str(v))
                self._updating = False

    def setValue(self, v):
        if v != self.lastValue:
            self.sld.setValue(int(v) if v else 0)
            self.changed(int(v) if v else 0)
            self.lastValue = v

    def setProperty():
        pass

    # MODEL / VIEW

    def setColumn(self, column):
        self._column = column

    def setModel(self, model):
        self._model = model
        self._model.dataChanged.connect(self.update)

    def update(self, topLeft, bottomRight):

        if self._updating:
            return

        if self._index:
            if topLeft.row() <= self._index.row() <= bottomRight.row():
                self.updateValue()

    def setCurrentModelIndex(self, index):
        if index.isValid():
            if index.column() != self._column:
                index = index.sibling(index.row(), self._column)
            self._index = index

            self.updateValue()

    def updateValue(self):
        if self._index:
            val = toInt(self._model.data(self._index))
            if self.sld.value() != val:
                self._updating = True
                self.setValue(val)
                self._updating = False

    importance = pyqtProperty(str, fget=getImportance, fset=setValue, notify=importanceChanged)



#!/usr/bin/env python
# --!-- coding: utf8 --!--

"""
The converters package provide functions to quickly convert on the fly from
one format to another. It is responsible to check what external library are
present, and do the job as best as possible with what we have in hand.
"""

from manuskript.converters.abstractConverter import abstractConverter
from manuskript.converters.pandocConverter import pandocConverter
#from manuskript.converters.markdownConverter import markdownConverter


def HTML2MD(html):

    # Convert using pandoc
    if pandocConverter.isValid():
        return pandocConverter.convert(html, _from="html", to="markdown")

    # Convert to plain text using QTextEdit
    return HTML2PlainText(html)


def HTML2PlainText(html):
    """
    Convert from HTML to plain text.
    """

    if pandocConverter.isValid():
        return pandocConverter.convert(html, _from="html", to="plain")

    # Last resort: probably resource inefficient
    e = QTextEdit()
    e.setHtml(html)
    return e.toPlainText()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/revisions_ui.ui'
#
# Created: Mon Oct 16 10:36:02 2017
#      by: PyQt5 UI code generator 5.2.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_revisions(object):
    def setupUi(self, revisions):
        revisions.setObjectName("revisions")
        revisions.resize(400, 344)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(revisions)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.splitter = QtWidgets.QSplitter(revisions)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.list = QtWidgets.QListWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.list.sizePolicy().hasHeightForWidth())
        self.list.setSizePolicy(sizePolicy)
        self.list.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.list.setObjectName("list")
        self.scrollArea = QtWidgets.QScrollArea(self.splitter)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 400, 68))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.view = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.view.sizePolicy().hasHeightForWidth())
        self.view.setSizePolicy(sizePolicy)
        self.view.setText("")
        self.view.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.view.setWordWrap(True)
        self.view.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.view.setObjectName("view")
        self.verticalLayout.addWidget(self.view)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.layoutWidget)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.btnOptions = QtWidgets.QPushButton(self.layoutWidget)
        icon = QtGui.QIcon.fromTheme("preferences-system-symbolic")
        self.btnOptions.setIcon(icon)
        self.btnOptions.setFlat(True)
        self.btnOptions.setObjectName("btnOptions")
        self.horizontalLayout_2.addWidget(self.btnOptions)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.btnRestore = QtWidgets.QPushButton(self.layoutWidget)
        icon = QtGui.QIcon.fromTheme("edit-redo")
        self.btnRestore.setIcon(icon)
        self.btnRestore.setFlat(True)
        self.btnRestore.setObjectName("btnRestore")
        self.horizontalLayout_2.addWidget(self.btnRestore)
        self.btnDelete = QtWidgets.QPushButton(self.layoutWidget)
        icon = QtGui.QIcon.fromTheme("edit-delete")
        self.btnDelete.setIcon(icon)
        self.btnDelete.setFlat(True)
        self.btnDelete.setObjectName("btnDelete")
        self.horizontalLayout_2.addWidget(self.btnDelete)
        self.verticalLayout_2.addWidget(self.splitter)

        self.retranslateUi(revisions)
        QtCore.QMetaObject.connectSlotsByName(revisions)

    def retranslateUi(self, revisions):
        _translate = QtCore.QCoreApplication.translate
        revisions.setWindowTitle(_translate("revisions", "Form"))
        self.btnOptions.setText(_translate("revisions", "Options"))
        self.btnRestore.setText(_translate("revisions", "Restore"))
        self.btnDelete.setText(_translate("revisions", "Delete"))




#!/usr/bin/env python
# --!-- coding: utf8 --!--

"""Tests for settingsWindow"""

def test_loadImportWiget(MWSampleProject):
    """
    Simply tests that import widget loads properly.
    """
    MW = MWSampleProject

    # Loading from mainWindow
    MW.doImport()
    I = MW.dialog
    assert I.isVisible()
    I.hide()
    I.close()

#FIXME: test significant stuff



#!/usr/bin/env python
# --!-- coding: utf8 --!--

from manuskript.exporter.manuskript import manuskriptExporter
from manuskript.exporter.pandoc import pandocExporter

exporters = [
    manuskriptExporter(),
    pandocExporter()
]

def getExporterByName(name):
    for e in exporters:
        if e.name == name:
            return e

    return None



#!/usr/bin/python
# -*- coding: utf8 -*-
from PyQt5.QtGui import QTextBlockUserData


class blockUserData(QTextBlockUserData):
    @staticmethod
    def getUserData(block):
        """Returns userData if it exists, or a blank one."""
        data = block.userData()
        if data is None:
            data = blockUserData()
        return data

    @staticmethod
    def getUserState(block):
        """Returns the block state."""
        state = block.userState()
        while state >= 100:
            state -= 100  # +100 means in a list
        return state

    def __init__(self):
        QTextBlockUserData.__init__(self)
        self._listLevel = 0
        self._leadingSpaces = 0
        self._emptyLinesBefore = 0
        self._listSymbol = ""

    def isList(self):
        return self._listLevel > 0

    def listLevel(self):
        return self._listLevel

    def setListLevel(self, level):
        self._listLevel = level

    def listSymbol(self):
        return self._listSymbol

    def setListSymbol(self, s):
        self._listSymbol = s

    def leadingSpaces(self):
        return self._leadingSpaces

    def setLeadingSpaces(self, n):
        self._leadingSpaces = n

    def emptyLinesBefore(self):
        return self._emptyLinesBefore

    def setEmptyLinesBefore(self, n):
        self._emptyLinesBefore = n

    def text(self):
        return str(self.listLevel()) + "|" + str(self.leadingSpaces()) + "|" + str(self.emptyLinesBefore())

    def __eq__(self, b):
        return self._listLevel == b._listLevel and \
               self._leadingSpaces == b._leadingSpaces and \
               self._emptyLinesBefore == b._emptyLinesBefore

    def __ne__(self, b):
        return not self == b



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/search_ui.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_search(object):
    def setupUi(self, search):
        search.setObjectName("search")
        search.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(search)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.text = QtWidgets.QLineEdit(search)
        self.text.setInputMask("")
        self.text.setFrame(False)
        self.text.setClearButtonEnabled(True)
        self.text.setObjectName("text")
        self.horizontalLayout.addWidget(self.text)
        self.btnOptions = QtWidgets.QPushButton(search)
        self.btnOptions.setText("")
        icon = QtGui.QIcon.fromTheme("edit-find")
        self.btnOptions.setIcon(icon)
        self.btnOptions.setCheckable(True)
        self.btnOptions.setFlat(True)
        self.btnOptions.setObjectName("btnOptions")
        self.horizontalLayout.addWidget(self.btnOptions)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.result = QtWidgets.QListWidget(search)
        self.result.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.result.setObjectName("result")
        self.verticalLayout.addWidget(self.result)

        self.retranslateUi(search)
        QtCore.QMetaObject.connectSlotsByName(search)

    def retranslateUi(self, search):
        _translate = QtCore.QCoreApplication.translate
        search.setWindowTitle(_translate("search", "Form"))
        self.text.setPlaceholderText(_translate("search", "Search for..."))




#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import QMutex
from PyQt5.QtWidgets import QLineEdit

from manuskript.enums import Outline
from manuskript.functions import toString


class lineEditView(QLineEdit):
    def __init__(self, parent=None):
        QLineEdit.__init__(self, parent)
        self._column = Outline.title
        self._indexes = None
        self._index = None
        self._placeholderText = None
        self._updating = QMutex()

    def setModel(self, model):
        self._model = model
        self._model.dataChanged.connect(self.update)

    def setColumn(self, col):
        self._column = col

    def setCurrentModelIndex(self, index):
        self._indexes = None
        if index.isValid():
            if index.column() != self._column:
                index = index.sibling(index.row(), self._column)
            self._index = index
            self._model = index.model()
            # self.item = index.internalPointer()
            if self._placeholderText is not None:
                self.setPlaceholderText(self._placeholderText)
            self.textEdited.connect(self.submit)
            self.updateText()

    def setCurrentModelIndexes(self, indexes):
        self._indexes = []
        self._index = None

        for i in indexes:
            if i.isValid():
                if i.column() != self._column:
                    i = i.sibling(i.row(), self._column)
                self._indexes.append(i)

        self.textEdited.connect(self.submit)
        self.updateText()

    def submit(self):
        self._updating.lock()
        text = self.text()
        self._updating.unlock()

        if self._index:
            # item = self._index.internalPointer()
            if text != self._model.data(self._index):
                self._model.setData(self._index, text)

        elif self._indexes:
            for i in self._indexes:
                # item = i.internalPointer()
                if text != self._model.data(i):
                    self._model.setData(i, text)

    def update(self, topLeft, bottomRight):
        update = False

        if self._index:
            if topLeft.row() <= self._index.row() <= bottomRight.row():
                update = True

        elif self._indexes:
            for i in self._indexes:
                if topLeft.row() <= i.row() <= bottomRight.row():
                    update = True

        if update:
            self.updateText()

    def updateText(self):
        self._updating.lock()

        if self._index:
            # item = self._index.internalPointer()
            # txt = toString(item.data(self._column))
            txt = toString(self._model.data(self._index))
            if self.text() != txt:
                self.setText(txt)

        elif self._indexes:
            t = []
            same = True
            for i in self._indexes:
                # item = i.internalPointer()
                # t.append(toString(item.data(self._column)))
                t.append(toString(self._model.data(i)))

            for t2 in t[1:]:
                if t2 != t[0]:
                    same = False
                    break

            if same:
                self.setText(t[0])
            else:
                self.setText("")

                if not self._placeholderText:
                    self._placeholderText = self.placeholderText()

                self.setPlaceholderText(self.tr("Various"))

        self._updating.unlock()




#!/usr/bin/env python
# --!-- coding: utf8 --!--
import os
from collections import OrderedDict

from PyQt5.QtCore import QSize
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QWidget, QListWidgetItem, QFileDialog

from manuskript import exporter
from manuskript.ui.exporters.exportersManager_ui import Ui_ExportersManager
from manuskript.ui import style as S


class exportersManager(QWidget, Ui_ExportersManager):

    exportersMightHaveChanged = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)
        self.lblExporterName.setStyleSheet(S.titleLabelSS())

        # Var
        self.currentExporter = None

        # Populates lite
        self.lstExporters.clear()
        for E in exporter.exporters:
            item = QListWidgetItem(QIcon(E.icon), E.name)
            self.lstExporters.addItem(item)

        # UI
        for i in range(self.lstExporters.count()):
            item = self.lstExporters.item(i)
            item.setSizeHint(QSize(item.sizeHint().width(), 42))
            item.setTextAlignment(Qt.AlignCenter)
        self.lstExporters.setMaximumWidth(150)
        self.lstExporters.setMinimumWidth(150)

        self.lstExporters.currentTextChanged.connect(self.updateUi)
        self.lstExportTo.currentTextChanged.connect(self.updateFormatDescription)

        self.lstExporters.setCurrentRow(0)

        self.btnSetPath.clicked.connect(self.setAppPath)
        self.txtPath.editingFinished.connect(self.updateAppPath)

    def updateUi(self, name):
        E = exporter.getExporterByName(name)
        self.currentExporter = E

        if not E:
            self.stack.setEnabled(False)
            return

        self.stack.setEnabled(True)

        # Updates name and description
        self.lblExporterName.setText(E.name)
        self.lblExporterDescription.setText(E.description)

        # Updates formats
        self.lstExportTo.clear()

        for f in E.exportTo:
            item = QListWidgetItem(QIcon.fromTheme(f.icon), f.name)
            self.lstExportTo.addItem(item)

        self.grpExportTo.layout().setStretch(0, 4)
        self.grpExportTo.layout().setStretch(1, 6)

        self.lstExportTo.setCurrentRow(0)

        # Updates path & version
        self.grpPath.setVisible(E.name != "Manuskript")  # We hide if exporter is manuskript

        # Installed
        if E.isValid() == 2:
            self.lblStatus.setText(self.tr("Installed"))
            self.lblStatus.setStyleSheet("color: darkGreen;")
            self.lblHelpText.setVisible(False)
            self.lblVersion.setVisible(True)
            self.lblVersionName.setVisible(True)
        elif E.isValid() == 1:
            self.lblStatus.setText(self.tr("Custom"))
            self.lblStatus.setStyleSheet("color: darkOrange;")
            self.lblHelpText.setVisible(False)
            self.lblVersion.setVisible(True)
            self.lblVersionName.setVisible(True)
        else:
            self.lblStatus.setText(self.tr("Not found"))
            self.lblStatus.setStyleSheet("color: red;")
            self.lblHelpText.setVisible(True)
            self.lblHelpText.setText(self.tr("{} not found. Install it, or set path manually.").format(name))
            self.lblVersion.setVisible(False)
            self.lblVersionName.setVisible(False)

        # Version
        self.lblVersion.setText(E.version())

        # Path
        if E.path():
            self.txtPath.setText(E.path())
        else:
            self.txtPath.setText(E.customPath)

    def updateFormatDescription(self, name):
        if self.currentExporter:
            f = self.currentExporter.getFormatByName(name)

            if not f:
                self.lblExportToDescription.setText("")

            else:

                desc = "<b>{}:</b> {}".format(
                    name,
                    f.description)

                if not f.isValid():
                    desc += "<br><br>" + \
                            self.tr("<b>Status:</b> uninstalled.") + \
                            "<br><br>" + \
                            self.tr("<b>Requires:</b> ") + f.InvalidBecause

                self.lblExportToDescription.setText(desc)

    def setAppPath(self):
        if self.currentExporter:
            E = self.currentExporter
            fn = QFileDialog.getOpenFileName(self,
                                             caption=self.tr("Set {} executable path.").format(E.cmd),
                                             directory=E.customPath)
            if fn[0]:
                self.updateAppPath(fn[0])

    def updateAppPath(self, path=""):
        if not path:
            path = self.txtPath.text()

        if self.currentExporter:
            E = self.currentExporter
            E.setCustomPath(path)
            self.txtPath.setText(E.customPath)

            self.updateUi(E.name)
            self.exportersMightHaveChanged.emit()




#!/usr/bin/env python
#--!-- coding: utf8 --!--

import os
import re
from random import *

from PyQt5.QtCore import Qt, QRect, QStandardPaths, QObject, QRegExp, QDir
from PyQt5.QtCore import QUrl, QTimer
from PyQt5.QtGui import QBrush, QIcon, QPainter, QColor, QImage, QPixmap
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import qApp, QFileDialog, QTextEdit

from manuskript.enums import Outline

# Used to detect multiple connections
AUC = Qt.AutoConnection | Qt.UniqueConnection
MW = None


def wordCount(text):
    t = text.strip().replace(" ", "\n").split("\n")
    t = [l for l in t if l]
    return len(t)

validate_ok = lambda *args, **kwargs: True
def uiParse(input, default, converter, validator=validate_ok):
    """
    uiParse is a utility function that intends to make it easy to convert
    user input to data that falls in the range of expected values the
    program is expecting to handle.

    It swallows all exceptions that happen during conversion.
    The validator should return True to permit the converted value.
    """
    result = default
    try:
        result = converter(input)
    except:
        pass  # failed to convert

    # Whitelist default value in case default type differs from converter output.
    if (result != default) and not validator(result):
        result = default
    return result


def toInt(text):
    if text:
        try:
            return int(text)
        except ValueError:
            pass

    return 0


def toFloat(text):
    if text:
        return float(text)
    else:
        return 0.


def toString(text):
    if text in [None, "None"]:
        return ""
    else:
        return str(text)


def drawProgress(painter, rect, progress, radius=0):
    from manuskript.ui import style as S
    progress = toFloat(progress)  # handle invalid input (issue #561)
    painter.setPen(Qt.NoPen)
    painter.setBrush(QColor(S.base)) # "#dddddd"
    painter.drawRoundedRect(rect, radius, radius)

    painter.setBrush(QBrush(colorFromProgress(progress)))

    r2 = QRect(rect)
    r2.setWidth(r2.width() * min(progress, 1))
    painter.drawRoundedRect(r2, radius, radius)


def colorFromProgress(progress):
    progress = toFloat(progress)
    c1 = QColor(Qt.red)
    c2 = QColor(Qt.blue)
    c3 = QColor(Qt.darkGreen)
    c4 = QColor("#FFA500")

    if progress < 0.3:
        return c1
    elif progress < 0.8:
        return c2
    elif progress > 1.2:
        return c4
    else:
        return c3


def mainWindow():
    global MW
    if not MW:
        for i in qApp.topLevelWidgets():
            if i.objectName() == "MainWindow":
                MW = i
                return MW
        return None
    else:
        return MW


def iconColor(icon):
    """Returns a QRgb from a QIcon, assuming its all the same color"""
    px = icon.pixmap(5, 5)
    if px.width() != 0:
        return QColor(QImage(px).pixel(2, 2))
    else:
        return QColor(Qt.transparent)


def iconFromColor(color):
    px = QPixmap(32, 32)
    px.fill(color)
    return QIcon(px)


def iconFromColorString(string):
    return iconFromColor(QColor(string))


def themeIcon(name):
    "Returns an icon for the given name."

    db = {
        "character":    "stock_people",
        "characters":   "stock_people",
        "plot":         "stock_shuffle",
        "plots":        "stock_shuffle",
        "world":        "emblem-web", #stock_timezone applications-internet
        "outline":      "gtk-index", #applications-versioncontrol
        "label":        "folder_color_picker",
        "status":       "applications-development",
        "text":         "view-text",
        "card":         "view-card",
        "outline":      "view-outline",
        "tree":         "view-list-tree",
        "spelling":     "tools-check-spelling"
    }

    if name in db:
        return QIcon.fromTheme(db[name])
    else:
        return QIcon()

def randomColor(mix=None):
    """Generates a random color. If mix (QColor) is given, mixes the random color and mix."""
    r = randint(0, 255)
    g = randint(0, 255)
    b = randint(0, 255)

    if mix:
        r = (r + mix.red()) / 2
        g = (g + mix.green()) / 2
        b = (b + mix.blue()) / 2

    return QColor(r, g, b)


def mixColors(col1, col2, f=.5):
    fromString = False
    if type(col1) == str:
        fromString = True
        col1 = QColor(col1)
    if type(col2) == str:
        col2 = QColor(col2)
    f2 = 1-f
    r = col1.red() * f + col2.red() * f2
    g = col1.green() * f + col2.green() * f2
    b = col1.blue() * f + col2.blue() * f2

    return QColor(r, g, b) if not fromString else QColor(r, g, b).name()


def outlineItemColors(item):

    from manuskript.ui import style as S

    """Takes an OutlineItem and returns a dict of colors."""
    colors = {}
    mw = mainWindow()

    # POV
    colors["POV"] = QColor(Qt.transparent)
    POV = item.data(Outline.POV)
    if POV == "":
        col = QColor(Qt.transparent)
    else:
        for i in range(mw.mdlCharacter.rowCount()):
            if mw.mdlCharacter.ID(i) == POV:
                colors["POV"] = iconColor(mw.mdlCharacter.icon(i))

    # Label
    lbl = item.data(Outline.label)
    if lbl == "":
        col = QColor(Qt.transparent)
    else:
        col = iconColor(mw.mdlLabels.item(toInt(lbl)).icon())
    # if col == Qt.black:
    #     # Don't know why, but transparent is rendered as black
    #     col = QColor(Qt.transparent)
    colors["Label"] = col

    # Progress
    pg = item.data(Outline.goalPercentage)
    colors["Progress"] = colorFromProgress(pg)

    # Compile
    if item.compile() in [0, "0"]:
        colors["Compile"] = mixColors(QColor(S.text), QColor(S.window))
    else:
        colors["Compile"] = QColor(Qt.transparent) # will use default

    return colors


def colorifyPixmap(pixmap, color):
    # FIXME: ugly
    p = QPainter(pixmap)
    p.setCompositionMode(p.CompositionMode_Overlay)
    p.fillRect(pixmap.rect(), color)
    return pixmap


def appPath(suffix=None):
    p = os.path.realpath(os.path.join(os.path.split(__file__)[0], "../.."))
    if suffix:
        p = os.path.join(p, suffix)
    return p


def writablePath(suffix=None):
    if hasattr(QStandardPaths, "AppLocalDataLocation"):
        p = QStandardPaths.writableLocation(QStandardPaths.AppLocalDataLocation)
    else:
        # Qt < 5.4
        p = QStandardPaths.writableLocation(QStandardPaths.DataLocation)
    if suffix:
        p = os.path.join(p, suffix)
    if not os.path.exists(p):
        os.makedirs(p)
    return p


def allPaths(suffix=None):
    paths = []
    # src directory
    paths.append(appPath(suffix))
    # user writable directory
    paths.append(writablePath(suffix))
    return paths

def tempFile(name):
    "Returns a temp file."
    return os.path.join(QDir.tempPath(), name)


def totalObjects():
    return len(mainWindow().findChildren(QObject))


def printObjects():
    print("Objects:", str(totalObjects()))


def findWidgetsOfClass(cls):
    """
    Returns all widgets, children of MainWindow, whose class is cls.
    @param cls: a class
    @return: list of QWidgets
    """
    return mainWindow().findChildren(cls, QRegExp())


def findBackground(filename):
    """
    Returns the full path to a background file of name filename within resources folders.
    """
    return findFirstFile(re.escape(filename), "resources/backgrounds")


def findFirstFile(regex, path="resources"):
    """
    Returns full path of first file matching regular expression regex within folder path,
    otherwise returns full path of last file in folder path.
    """
    paths = allPaths(path)
    for p in paths:
        lst = os.listdir(p)
        for l in lst:
            if re.match(regex, l):
                return os.path.join(p, l)

def customIcons():
    """
    Returns a list of possible customIcons. String from theme.
    """

    r = [
        "text-plain",
        "gnome-settings",
        "applications-internet",
        "applications-debugging",
        "applications-development",
        "system-help",
        "info",
        "dialog-question",
        "dialog-warning",
        "stock_timezone",
        "stock_people",
        "stock_shuffle",
        "gtk-index",
        "folder_color_picker",
        "applications-versioncontrol",

        "stock_home",
        "stock_trash_empty",
        "stock_trash_full",
        "stock_yes",
        "stock_no",
        "stock_notes",
        "stock_calendar",
        "stock_mic",
        'stock_score-lowest', 'stock_score-lower', 'stock_score-low', 'stock_score-normal', 'stock_score-high', 'stock_score-higher', 'stock_score-highest',
        "stock_task",
        "stock_refresh",
        "application-community",
        "applications-chat",
        "application-menu",
        "applications-education",
        "applications-science",
        "applications-puzzles",
        "applications-roleplaying",
        "applications-sports",
        "applications-libraries",
        "applications-publishing",
        "applications-development",
        "applications-games",
        "applications-boardgames",
        "applications-geography",
        "applications-physics",
        "package_multimedia",
        "media-flash",
        "media-optical",
        "media-floppy",
        "media-playback-start",
        "media-playback-pause",
        "media-playback-stop",
        "media-playback-record",
        "media-playback-start-rtl",
        "media-eject",
        "document-save",
        "gohome",
        'purple-folder', 'yellow-folder', 'red-folder', 'custom-folder', 'grey-folder', 'blue-folder', 'default-folder', 'pink-folder', 'orange-folder', 'green-folder', 'brown-folder',
        'folder-home', 'folder-remote', 'folder-music', 'folder-saved-search', 'folder-projects', 'folder-sound', 'folder-publicshare', 'folder-pictures', 'folder-saved-search-alt', 'folder-tag',
        'calendar-01', 'calendar-02', 'calendar-03', 'calendar-04', 'calendar-05', 'calendar-06', 'calendar-07', 'calendar-08', 'calendar-09', 'calendar-10',
        'arrow-down', 'arrow-left', 'arrow-right', 'arrow-up', 'arrow-down-double', 'arrow-left-double', 'arrow-right-double', 'arrow-up-double',
        'emblem-added', 'emblem-checked', 'emblem-downloads', 'emblem-dropbox-syncing', 'emblem-danger', 'emblem-development', 'emblem-dropbox-app', 'emblem-art', 'emblem-camera', 'emblem-dropbox-selsync', 'emblem-insync-des-error', 'emblem-insync-error', 'emblem-generic', 'emblem-favorites', 'emblem-error', 'emblem-dropbox-uptodate', 'emblem-marketing', 'emblem-money', 'emblem-music', 'emblem-noread', 'emblem-people', 'emblem-personal', 'emblem-sound', 'emblem-shared', 'emblem-sales', 'emblem-presentation', 'emblem-plan', 'emblem-system', 'emblem-urgent', 'emblem-videos', 'emblem-web',
        'face-angel', 'face-clown', 'face-angry', 'face-cool', 'face-devilish', 'face-sick', 'face-sleeping', 'face-uncertain', 'face-monkey', 'face-ninja', 'face-pirate', 'face-glasses', 'face-in-love', 'face-confused',
        'feed-marked-symbolic', 'feed-non-starred', 'feed-starred', 'feed-unmarked-symbolic',
        'notification-new-symbolic',
        ]

    return sorted(r)

def statusMessage(message, duration=5000, importance=1):
    """
    Shows a message in MainWindow's status bar.
    Importance: 0 = low, 1 = normal, 2 = important, 3 = critical.
    """
    from manuskript.ui import style as S
    MW.statusBar().hide()
    MW.statusLabel.setText(message)
    if importance == 0:
        MW.statusLabel.setStyleSheet("color:{};".format(S.textLighter))
    elif importance == 1:
        MW.statusLabel.setStyleSheet("color:{};".format(S.textLight))
    elif importance == 2:
        MW.statusLabel.setStyleSheet("color:{}; font-weight: bold;".format(S.text))
    elif importance == 3:
        MW.statusLabel.setStyleSheet("color:red; font-weight: bold;")
    MW.statusLabel.adjustSize()
    g = MW.statusLabel.geometry()
    # g.moveCenter(MW.mapFromGlobal(MW.geometry().center()))
    s = MW.layout().spacing() / 2
    g.setLeft(s)
    g.moveBottom(MW.mapFromGlobal(MW.geometry().bottomLeft()).y() - s)
    MW.statusLabel.setGeometry(g)
    MW.statusLabel.show()
    QTimer.singleShot(duration, MW.statusLabel.hide)

def openURL(url):
    """
    Opens url (string) in browser using desktop default application.
    """
    QDesktopServices.openUrl(QUrl(url))

def getSaveFileNameWithSuffix(parent, caption, directory, filter, options=None, selectedFilter=None, defaultSuffix=None):
    """
    A reimplemented version of QFileDialog.getSaveFileName() because we would like to make use
    of the QFileDialog.defaultSuffix property that getSaveFileName() does not let us adjust.

    Note: knowing the selected filter is not an invitation to change the chosen filename later.
    """
    dialog = QFileDialog(parent=parent, caption=caption, directory=directory, filter=filter)
    if options:
        dialog.setOptions(options)
    if defaultSuffix:
        dialog.setDefaultSuffix(defaultSuffix)
    dialog.setFileMode(QFileDialog.AnyFile)
    if hasattr(dialog, 'setSupportedSchemes'): # Pre-Qt5.6 lacks this.
        dialog.setSupportedSchemes(("file",))
    dialog.setAcceptMode(QFileDialog.AcceptSave)
    if selectedFilter:
        dialog.selectNameFilter(selectedFilter)
    if (dialog.exec() == QFileDialog.Accepted):
        return dialog.selectedFiles()[0], dialog.selectedNameFilter()
    return None, None

def inspect():
    """
    Debugging tool. Call it to see a stack of calls up to that point.
    """
    import inspect, os
    print("-----------------------")
    for s in inspect.stack()[1:]:
        print(" * {}:{} // {}".format(
            os.path.basename(s.filename),
            s.lineno,
            s.function))
        print("   " + "".join(s.code_context))

# Spellchecker loads writablePath from this file, so we need to load it after they get defined
from manuskript.functions.spellchecker import Spellchecker



#!/usr/bin/env python
# --!-- coding: utf8 --!--

import locale

from PyQt5.QtCore import QAbstractItemModel, QMimeData
from PyQt5.QtCore import QModelIndex
from PyQt5.QtCore import QSize
from PyQt5.QtCore import QVariant
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QFont
from PyQt5.QtWidgets import QTextEdit, qApp

from manuskript import settings
from lxml import etree as ET

from manuskript.enums import Outline
from manuskript.functions import mainWindow, toInt, wordCount
from manuskript.models import outlineItem

try:
    locale.setlocale(locale.LC_ALL, '')
except:
    # Invalid locale, but not really a big deal because it's used only for
    # number formatting
    pass
import time, os


class abstractModel(QAbstractItemModel):
    """
    Abstract model is the base class for all others models we use.

    It's main responsibilities are:

    - Interface with QModelIndex and stuff
    - XML Import / Export
    - Drag'n'drop

    """
    def __init__(self, parent):
        QAbstractItemModel.__init__(self, parent)

        self.rootItem = outlineItem(self, title="Root", ID="0")

        # Stores removed item, in order to remove them on disk when saving, depending on the file format.
        self.removed = []
        self._removingRows = False

    def index(self, row, column, parent):

        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()

    def indexFromItem(self, item, column=0):
        if item == self.rootItem:
            return QModelIndex()

        parent = item.parent()
        if not parent:
            parent = self.rootItem

        if len(parent.children()) == 0:
            return None

        # print(item.title(), [i.title() for i in parent.children()])

        row = parent.children().index(item)
        col = column
        return self.createIndex(row, col, item)

    def ID(self, index):
        if index.isValid():
            item = index.internalPointer()
            return item.ID()

    def findItemsContaining(self, text, columns, caseSensitive=False):
        """
        Returns a list of IDs of all items containing `text`
        in columns `columns` (being a list of int).
        """
        return self.rootItem.findItemsContaining(text, columns, mainWindow(), caseSensitive)

    def getItemByID(self, ID, ignore=None):
        """Returns the item whose ID is `ID`, unless this item matches `ignore`."""

        def search(item):
            if item.ID() == ID:
                if item == ignore:
                    # The item we really want won't be found in the children of this
                    # particular item anymore; stop searching this branch entirely.
                    return None
                return item
            for c in item.children():
                r = search(c)
                if r:
                    return r

        item = search(self.rootItem)
        return item

    def getIndexByID(self, ID, column=0, ignore=None):
        """Returns the index of item whose ID is `ID`. If none, returns QModelIndex().

        If `ignore` is set, it will not return that item if found as valid match for the ID"""

        item = self.getItemByID(ID, ignore=ignore)
        if not item:
            return QModelIndex()
        else:
            return self.indexFromItem(item, column)

    def parent(self, index=QModelIndex()):
        if not index.isValid():
            return QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        # Check whether the parent is the root, or is otherwise invalid.
        # That is to say: no parent or the parent lacks a parent.
        if (parentItem == self.rootItem) or \
           (parentItem is None) or (parentItem.parent() is None):
            return QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent=QModelIndex()):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        return parentItem.childCount()

    def columnCount(self, parent=QModelIndex()):
        if parent.isValid():
            return parent.internalPointer().columnCount()
        else:
            return self.rootItem.columnCount()

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return QVariant()

        item = index.internalPointer()
        return item.data(index.column(), role)

    def setData(self, index, value, role=Qt.EditRole):
        item = index.internalPointer()
        if item.data(index.column(), role) != value:

            item.setData(index.column(), value, role)

            # self.dataChanged.emit(index.sibling(index.row(), 0),
            # index.sibling(index.row(), max([i.value for i in Outline])))
            # print("Model emit", index.row(), index.column())
            self.dataChanged.emit(index, index)

            if index.column() == Outline.type:
                # If type changed, then the icon of title changed.
                # Some views might be glad to know it.
                self.dataChanged.emit(index.sibling(index.row(), Outline.title),
                                      index.sibling(index.row(), Outline.title))

        return True

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role in [Qt.DisplayRole, Qt.ToolTipRole]:
            if section == Outline.title:
                return self.tr("Title")
            elif section == Outline.POV:
                return self.tr("POV")
            elif section == Outline.label:
                return self.tr("Label")
            elif section == Outline.status:
                return self.tr("Status")
            elif section == Outline.compile:
                return self.tr("Compile")
            elif section == Outline.wordCount:
                return self.tr("Word count")
            elif section == Outline.goal:
                return self.tr("Goal")
            elif section == Outline.goalPercentage:
                return "%"
            else:
                return [i.name for i in Outline][section]

        elif role == Qt.SizeHintRole:
            if section == Outline.compile:
                return QSize(40, 30)
            elif section == Outline.goalPercentage:
                return QSize(100, 30)
            else:
                return QVariant()
        else:
            return QVariant()

        return True

    def maxLevel(self):
        """Returns the max depth of the model."""
        def depth(item, d=-1):
            d += 1
            r = d
            for c in item.children():
                r = max(r, depth(c, d))
            return r

        d = depth(self.rootItem)
        return d

    #################### DRAG AND DROP ########################
    # http://doc.qt.io/qt-5/model-view-programming.html#using-drag-and-drop-with-item-views

    def flags(self, index):
        # FIXME when dragging folders, sometimes flags is not called

        flags = QAbstractItemModel.flags(self, index) | Qt.ItemIsEditable

        if index.isValid() and index.internalPointer().isFolder() and index.column() == 0:
            flags |= Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled

        elif index.isValid() and index.column() == 0:
            flags |= Qt.ItemIsDragEnabled

        elif not index.isValid():
            flags |= Qt.ItemIsDropEnabled

        if index.isValid() and index.column() == Outline.compile:
            flags |= Qt.ItemIsUserCheckable

        if index.column() in [i.value for i in [Outline.wordCount, Outline.goalPercentage]]:
            flags &= ~ Qt.ItemIsEditable

        return flags

    def mimeTypes(self):
        return ["application/xml"]

    def mimeData(self, indexes):
        mimeData = QMimeData()
        encodedData = ""

        root = ET.Element("outlineItems")

        for index in indexes:
            if index.isValid() and index.column() == 0:
                item = ET.XML(index.internalPointer().toXML())
                root.append(item)

        encodedData = ET.tostring(root)

        mimeData.setData("application/xml", encodedData)
        return mimeData

    def supportedDropActions(self):

        return Qt.CopyAction | Qt.MoveAction

    def canDropMimeData(self, data, action, row, column, parent):
        """Ensures that we are not dropping an item into itself."""

        if not data.hasFormat("application/xml"):
            return False

        if column > 0:
            return False

        # # Gets encoded mime data to retrieve the item
        items = self.decodeMimeData(data)
        if items is None:
            return False

        # We check if parent is not a child of one of the items
        if self.isParentAChildOfItems(parent, items):
            return False

        return True

    def isParentAChildOfItems(self, parent, items):
        """
        Takes a parent index, and a list of outlineItems items. Check whether
        parent is in a child of one of the items.
        Return True in that case, False if not.
        """

        # Get the parent item
        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        for item in items:
            # Get parentItem's parents IDs in a list
            path = parentItem.pathID()  # path to item in the form [(ID, title), ...]
            path = [ID for ID, title in path]
            # Is item in the path? It would mean that it tries to get dropped
            # as a children of himself.
            if item.ID() in path:
                return True

        return False

    def decodeMimeData(self, data):
        if not data.hasFormat("application/xml"):
            return None
        encodedData = bytes(data.data("application/xml")).decode()
        root = ET.XML(encodedData)
        if root is None:
            return None

        if root.tag != "outlineItems":
            return None

        items = []
        for child in root:
            if child.tag == "outlineItem":
                item = outlineItem(xml=ET.tostring(child))
                items.append(item)

        # We remove every item whose parent is also in items, otherwise it gets
        # duplicated. (https://github.com/olivierkes/manuskript/issues/169)
        # For example if selecting:
        #   - Parent
        #      - Child
        # And dragging them, items encoded in mime data are: [Parent, Child],
        # but Child is already contained in Parent, so if we do nothing we end
        # up with:
        #   - Parent
        #     - Child
        #   - Child

        newItems = items[:]
        IDs = [i.ID() for i in items]

        def checkIfChildIsPresent(item):
            # Recursively check every children of item, to see if any is in
            # the list of items to copy. If so, we remove it from the list.
            for c in item.children():
                # We check if children is in the selection
                # and if it hasn't been removed yet
                if c.ID() in IDs and c.ID() in [i.ID() for i in newItems]:
                    # Remove item by ID
                    newItems.remove([i for i in newItems if i.ID() == c.ID()][0])
                checkIfChildIsPresent(c)

        for i in items:
            checkIfChildIsPresent(i)

        items = newItems

        return items

    def dropMimeData(self, data, action, row, column, parent):

        if action == Qt.IgnoreAction:
            return True  # What is that?

        if action == Qt.MoveAction:
            # Strangely, on some cases, we get a call to dropMimeData though
            # self.canDropMimeData returned False.
            # See https://github.com/olivierkes/manuskript/issues/169 to reproduce.
            # So we double check for safety.
            if not self.canDropMimeData(data, action, row, column, parent):
                return False

        items = self.decodeMimeData(data)

        if items is None:
            return False

        if column > 0:
            column = 0

        if row != -1:
            beginRow = row
        elif parent.isValid():
            beginRow = self.rowCount(parent) + 1
        else:
            beginRow = self.rowCount() + 1

        if action == Qt.CopyAction:
            # Behavior if parent is a text item
            # For example, we select a text and do: CTRL+C CTRL+V
            if parent.isValid() and not parent.internalPointer().isFolder():
                # We insert copy in parent folder, just below
                beginRow = parent.row() + 1
                parent = parent.parent()

            if parent.isValid() and parent.internalPointer().isFolder():
                while self.isParentAChildOfItems(parent, items):
                    # We are copying a folder on itself. Assume duplicates.
                    # Copy not in, but next to
                    beginRow = parent.row() + 1
                    parent = parent.parent()

        if not items:
            return False

        # In case of copy actions, items might be duplicates, so we need new IDs.
        # But they might not be, if we cut, then paste. Paste is a Copy Action.
        # The first paste would not need new IDs. But subsequent ones will.
        if action == Qt.CopyAction:
            IDs = self.rootItem.listAllIDs()

            for item in items:
                if item.ID() in IDs:
                    # Recursively remove ID. So will get a new one when inserted.
                    def stripID(item):
                        item.setData(Outline.ID, None)
                        for c in item.children():
                            stripID(c)

                    stripID(item)

        r = self.insertItems(items, beginRow, parent)

        return r

    ################# ADDING AND REMOVING #################

    def insertItem(self, item, row, parent=QModelIndex()):
        return self.insertItems([item], row, parent)

    def insertItems(self, items, row, parent=QModelIndex()):
        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        if parent.isValid() and parent.column() != 0:
            parent = parentItem.index()

        # Insert only if parent is folder
        if parentItem.isFolder():
            self.beginInsertRows(parent, row, row + len(items) - 1)

            for i in items:
                parentItem.insertChild(row + items.index(i), i)

            self.endInsertRows()

            return True

        else:
            return False

    def appendItem(self, item, parent=QModelIndex()):
        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        if parent.isValid() and parent.column() != 0:
            parent = parentItem.index()

        # If parent is folder, write into
        if parentItem.isFolder():
            self.insertItem(item, self.rowCount(parent), parent)

        # If parent is not folder, write next to
        else:
            self.insertItem(item, parent.row() + 1, parent.parent())

    def removeIndex(self, index):
        item = index.internalPointer()
        self.removeRow(item.row(), index.parent())

    def removeIndexes(self, indexes):
        levels = {}
        for i in indexes:
            item = i.internalPointer()
            level = item.level()
            if not level in levels:
                levels[level] = []
            levels[level].append([i.row(), i])

        # Sort by level then by row
        for l in reversed(sorted(levels.keys())):
            rows = levels[l]

            rows = list(reversed(sorted(rows, key=lambda x: x[0])))
            for r in rows:
                self.removeIndex(r[1])

    def removeRow(self, row, parent=QModelIndex()):
        return self.removeRows(row, 1, parent)

    def removeRows(self, row, count, parent=QModelIndex()):
        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        self._removingRows = True  # Views that are updating can easily know
                                   # if this is due to row removal.
        self.beginRemoveRows(parent, row, row + count - 1)
        for i in range(count):
            item = parentItem.removeChild(row)
            self.removed.append(item)

        self._removingRows = False
        self.endRemoveRows()
        return True

        # def insertRow(self, row, item, parent=QModelIndex()):
        # self.beginInsertRows(parent, row, row)

        # if not parent.isValid():
        # parentItem = self.rootItem
        # else:
        # parentItem = parent.internalPointer()

        # parentItem.insertChild(row, item)

        # self.endInsertRows()

    ################# XML / saving / loading #################

    def saveToXML(self, xml=None):
        "If xml (filename) is given, saves the items to xml. Otherwise returns as string."
        root = ET.XML(self.rootItem.toXML())
        if xml:
            ET.ElementTree(root).write(xml, encoding="UTF-8", xml_declaration=True, pretty_print=True)
        else:
            return ET.tostring(root, encoding="UTF-8", xml_declaration=True, pretty_print=True)

    def loadFromXML(self, xml, fromString=False):
        "Load from xml. Assume that xml is a filename. If fromString=True, xml is the content."
        if not fromString:
            root = ET.parse(xml)
        else:
            root = ET.fromstring(xml)

        self.rootItem = outlineItem(model=self, xml=ET.tostring(root), ID="0")
        self.rootItem.checkIDs()

    def indexFromPath(self, path):
        path = path.split(",")
        item = self.rootItem
        for p in path:
            if p != "" and int(p) < item.childCount():
                item = item.child(int(p))
        return self.indexFromItem(item)
